#!/user/bin/env python
'''containsLProteinChain.py

This filter returns entries that contain protein chain(s) made of L-amino acids.
The default constructor returns entries that contain at least one
polymer chain that is an L-protein. If the "exclusive" flag is set to true
in the constructor, all polymer chains must be L-proteins. For a multi-model structure,
this filter only checks the first model.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from mmtfPyspark.filters import ContainsPolymerChainType


class ContainsLProteinChain(object):
    '''Default constructor matches any entry that contains at least one
    L-protein chain. As an example a L-protein/DNA complex passes this filter

        Optional constructor that can be used to filter entries that exclusively
    contain L-protein chains. For example, with "exclusive" set to true, a
    L-protein/DNA complex does not pass this filter.

    Attributes
    ----------
        exclusive (bool): if true, only return entries that are exclusively
                          contain L-protein chains
    '''

    def __init__(self, exclusive=False):
        self.filter = ContainsPolymerChainType([
            ContainsPolymerChainType.L_PEPTIDE_LINKING,
            ContainsPolymerChainType.PEPTIDE_LINKING], exclusive=exclusive)

    def __call__(self, t):
        return self.filter(t)



#!/user/bin/env python
'''dataBalancer.py:

Creates a balanced dataset for classification problems by either
downsampling the majority classes or upsampling the  minority classes.
It randomly samples each class and returns a dataset with approximately
the same number of samples in each class

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from pyspark.sql import DataFrame
from pyspark.sql import Row
from functools import reduce
import math


def downsample(data, columnName, seed=7):
    '''Returns a balanced dataset for the given column name by downsampling
    the majority classes.
    The classification column must be of type String

    Attributes
    ----------
        data (Dataframe)
        columnName (str): column to be balanced by
        seed (int): random number seed
    '''

    counts = data.groupby(columnName).count().collect()

    count = [int(x[1]) for x in counts]
    names = [y[0] for y in counts]
    minCount = min(count)

    samples = [data.filter(columnName + "='%s'" % n)
               .sample(False, minCount / float(c), seed)
               for n, c in zip(names, count)]

    return reduce(lambda x, y: x.union(y), samples)


def upsample(data, columnName, seed=7):
    '''Returns a balanced dataset for the given column name by upsampling
    the majority classes.
    The classification column must be of type String

    Attributes
    ----------
        data (Dataframe)
        columnName  (str): column to be balanced by
        seed (int): random number seed
    '''

    counts = data.groupby(columnName).count().collect()

    count = [int(x[1]) for x in counts]
    names = [y[0] for y in counts]
    maxCount = max(count)

    samples = [data.filter(columnName + "='%s'" % n)
               .sample(False, maxCount / float(c), seed)
               if abs(1 - maxCount / float(c)) > 1.0
               else data.filter(columnName + "='%s'" % n)
               for n, c in zip(names, count)
               ]

    return reduce(lambda x, y: x.union(y), samples)



#!/user/bin/env python
'''groupInteractionExtractor.py:

Creates a dataset of interactions of a specified group within
a cutoff distance. Groups are specified by there
Chemical Component identifier (residue name), e.g., "ZN", "ATP".

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Obsolete"
'''

from mmtfPyspark.ml import pythonRDDToDataset
from mmtfPyspark.utils import StructureToAllInteractions

class groupInteractionExtractor(object):
    '''Class that creates a dataset of interactions of a specifed gorup within
    a cutoff distance

    Attributes:
        groupName (String): name of the group to be analyzed
        distance (float): cutoff distance
    '''

    def __init__(self, groupName, distance):
        self.groupName = groupName
        self.distance = distance


    def get_dataset(self, structures):
        '''Returns a dataset of residues that interact with specified group
        within a specified cutoff distance

        Attricutes:
            structure (pythonRdd): a set of PDB structures
        Returns:
            dataset with interacting residue and atom information
        '''
        # create a list of all residues with a threshold distance
        rows = structures.flatMap(StructureToAllInteractions(self.groupName, self.distance))

        # convert to a dataset
        colNames = ["structureId", "residue1", "atom1", "element1", "index1",
                    "residue2", "atom2", "element2", "index2", "distance"]
        return pythonRDDToDataset.get_dataset(rows, colNames)



#!/user/bin/env python
'''secondaryStructureExtractor.py

Creates a dataset of DSSP secondary structure assignments. The dataset
includes protein sequence, the DSSP 3-state (Q3) and 8-state (Q8)
assignments, and the fraction of alpha, beta, and coil within a chain. The
input to this class must be a single protein chain.

Example
-------
    get dataset of secondary structure assignments:

    pdb.flatMapToPair(new StructureToPolymerChains())
       .filter(new ContainsLProteinChain())
    secStruct = SecondaryStructureExtractor.getDataset(pdb)
    secStruct.show(10)

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from mmtfPyspark.ml import pythonRDDToDataset
from mmtfPyspark.utils import DsspSecondaryStructure
from pyspark.sql import Row


def get_dataset(structure):
    '''Returns a dataset with protein sequence and secondary structure assignments.

    Attributes
    ----------
        structure (mmtfStructure): single protein chain

    Returns
    -------
        dataset with sequence and secondary structure assignments
    '''

    rows = structure.map(
        lambda x: _get_sec_struct_fractions(x))  # Map or flatMap

    # convert to dataset
    colNames = ["structureChainId", "sequence", "alpha", "beta",
                "coil", "dsspQ8Code", "dsspQ3Code"]

    return pythonRDDToDataset.get_dataset(rows, colNames)


def get_python_rdd(structure):
    '''Returns a pythonRDD of 3-state secondary structure

    Attributes
    ----------
        structure (mmtfStructure)
    '''

    return structure.map(lambda x: _get_sec_struct_fractions(x))


def _get_sec_struct_fractions(t):
    '''Get factions of alpha, beta and coil within a chain
    '''

    key = t[0]
    structure = t[1]
    if structure.num_chains != 1:
        raise Exception(
            "This method can only be applied to single polyer chain.")

    dsspQ8, dsspQ3 = '', ''

    helix = 0
    sheet = 0
    coil = 0
    dsspIndex = 0
    structureIndex = 0

    for code in structure.sec_struct_list:

        seqIndex = structure.sequence_index_list[structureIndex]

        while dsspIndex < seqIndex:
            dsspQ3 += "X"
            dsspQ8 += "X"
            dsspIndex += 1

        structureIndex += 1
        dsspQ8 += DsspSecondaryStructure.get_dssp_code(
            code).get_one_letter_code()
        dsspIndex += 1

        q3 = DsspSecondaryStructure.get_q3_code(code).name
        if q3 == "ALPHA_HELIX":
            helix += 1
            dsspQ3 += "H"
        elif q3 == "EXTENDED":
            sheet += 1
            dsspQ3 += "E"
        elif q3 == "COIL":
            coil += 1
            dsspQ3 += "C"

    while dsspIndex < len(structure.entity_list[0]['sequence']):
        dsspQ8 += "X"
        dsspQ3 += "X"
        dsspIndex += 1

    n = len(structure.sec_struct_list)

    helix /= n
    sheet /= n
    coil /= n

    return Row(key, structure.entity_list[0]['sequence'], helix, sheet,
               coil, dsspQ8, dsspQ3)



#!/user/bin/env python
'''piscesDownloader.py

This class downloads representative protein chains from the PISCES
CulledPDB sets. A CulledPDB set is selected by specifying
sequenceIdentity and resolution cutoff values from the following
list:
    sequenceIdentity = [20, 25, 30, 40, 50, 60, 70, 80, 90]
    resolution = [1.6, 1.8, 2.0, 2.2, 2.5, 3.0]

References
----------
    PISCES:
        http://dunbrack.fccc.edu/PISCES.php
        Please cite the following in any work that uses lists provided by PISCES
        G. Wang and R. L. Dunbrack, Jr. PISCES: a protein sequence culling
        server. Bioinformatics, 19:1589-1591, 2003.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from urllib.request import urlopen
import gzip


class PiscesDownloader(object):
    '''
    Downloads representative protein chains from the PISCES
    CulledPDB sets. A CulledPDB set is selected by specifying
    sequenceIdentity and resolution cutoff values from the following lists:
    <p> sequenceIdentity = 20, 25, 30, 40, 50, 60, 70, 80, 90
    <p> resolution = 1.6, 1.8, 2.0, 2.2, 2.5, 3.0

    Attributes
    ----------
        sequenceIdentity (int): sequence identity [0]
        resoltion (int): resoltion for pisces filter [0.0]
    '''

    URL = "http://dunbrack.fccc.edu/Guoli/culledpdb_hh"
    SEQ_ID_LIST = [20, 25, 30, 40, 50, 60, 70, 80, 90]
    RESOLUTION_LIST = [1.6, 1.8, 2.0, 2.2, 2.5, 3.0]

    def __init__(self, sequenceIdentity=0, resolution=0.0):

        if sequenceIdentity not in self.SEQ_ID_LIST:
            raise ValueError("Invalid sequenceIdentity")

        if resolution not in self.RESOLUTION_LIST:
            raise ValueError("Invalid resolution value")

        self.sequenceIdentity = sequenceIdentity
        self.resolution = resolution

    def get_structure_chain_ids(self):
        fileURL = self.URL + '/' + self._get_file_name()
        u = urlopen(fileURL)
        line = str(gzip.GzipFile(fileobj=u).read()).split('\\n')
        structureChainId = [l.split()[0][:4] + '.' + l.split()[0][4] for l in line if
                            len(l.split()) > 1]
        return structureChainId

    def _get_file_name(self):
        u = urlopen(self.URL)
        fileName = ""
        cs = "pc" + str(self.sequenceIdentity) + "_res" + str(self.resolution)

        while True:

            line = str(u.readline())
            line = line.split('\"')[1].split('/')[-1]
            if line[:3] == 'log':
                break

            if (cs in line) and ("fasta" not in line):
                fileName = line
                break

        u.close()
        return fileName



#!/user/bin/env python
'''columnarStructureX.py

Inheritance class of ColumnarStructure

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''
import numpy as np
import sys
from mmtfPyspark.utils import ColumnarStructure
from sympy import Point3D


class ColumnarStructureX(ColumnarStructure):
    '''Inheritance of class ColumnarStructure with additional functions

    Attributes
    ----------
        structure (mmtfStructure): mmtf structure
        firstModelOnly (bool): flag to use only the first model of the structure
    '''

    def __init__(self, structure, firstModelOnly = True):

        ColumnarStructure.__init__(self, structure, firstModelOnly)
        self.normalizedbFactors = None
        self.clampedNormalizedbFactor = None


    def get_normalized_b_factors(self):
        '''Returns z-scores for B-factors (normalized B-factors).

        Critical z-score values: Confidence level Tail Area z critical
        90% 0.05 +- 1.645
        95% 0.025 +- 1.96
        99% 0.005 +- 2.576
        '''

        if self.normalizedbFactors is None:

            self.get_entity_types()
            self.bFactors = self.get_b_factors()
            self.entityTypes = self.get_entity_types()
            # Filter out DOD and HOH
            stats = np.array([self.bFactors[i] for i in range(self.get_num_atoms())\
                              if self.entityTypes[i] is not 'WAT'])
            # Define normalize function
            normalize = lambda x: (x - stats.mean()) / stats.std()
            if stats.std() != 0:
                self.normalizedbFactors = [float(n) for n in normalize(self.bFactors)]
            else:
                self.normalizedbFactors = [sys.float_info.max] * len(self.bFactors)

        return self.normalizedbFactors


    def get_clamped_normalized_b_factors(self):
        '''Returns a normalized B-factors that are clamped to the [-1,1] interval
        using the method of Liu et at. B-factors are normalized and scaled the
        90% Confidenceinterval of the B-factors to [-1,1]. Any value outside of
        the 90% confidence interval is set to either -1 or 1, whichever is closer.

        Reference
        ---------
            Liu et al. BMC Bioinformatics 2014, 15(Suppl 16):S3,
            Use B-factor related features for accurate classification between
            protein binding interfaces and crystal packing contacts
            <"https://doi.org/10.1186/1471-2105-15-S16-S3">
        '''

        if self.clampedNormalizedbFactor is None:

            self.get_normalized_b_factors()
            self.clampedNormalizedbFactor = self.normalizedbFactors.copy()

            # Normalize and scale the 90% confidence interval of the B factor to [-1,1]
            self.clampedNormalizedbFactor = self.clampedNormalizedbFactor / 1.645

            # Set any value outside the 90% interval to either -1 or 1
            self.clampedNormalizedbFactor[self.clampedNormalizedbFactor < -1.0] = -1.0
            self.clampedNormalizedbFactor[self.clampedNormalizedbFactor > 1.0] = 1.0

        return self.clampedNormalizedbFactor


    def get_calpha_coordinates(self):
        '''Get the coordinates for Calpha atoms
        '''

        self.get_calpha_atom_indices()

        x = self.get_x_coords()
        y = self.get_y_coords()
        z = self.get_z_coords()

        # TODO: Point3D extremely slow, only use if nessassary
        #calpha_coords_list = [Point3D(x[i], y[i], z[i]) for i in self.caIndices]
        calpha_coords_list = [np.array([x[i], y[i], z[i]]) for i in self.caIndices]
        self.calpha_coords = np.array(calpha_coords_list)

        return self.calpha_coords


    def get_calpha_atom_indices(self):
        '''Get the indices of Calpha atoms
        '''

        self.get_entity_types()
        self.get_atom_names()

        caIndices_list = [i for i in range(self.get_num_atoms()) \
                          if (self.atomNames[i] == "CA" \
                          and self.entityTypes[i] == "PRO")]

        self.caIndices = np.array(caIndices_list)

        return self.caIndices



#!/user/bin/env python
'''PdbjMineSearch.py

This filter runs an PDBj Mine 2 Search web service using an SQL query.

References
----------
Each category represents a table, and fields represent database columns, see
available tables and columns:
    https://pdbj.org/mine-rdb-docs

Data are provided through, Mine 2 SQL:
    https://pdbj.org/help/mine2-sql

Queries can be designed with the interactive PDBjMine2 query service:
    https://pdbj.org/mine/sql

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

import urllib
import tempfile
from pyspark.sql import SparkSession
from mmtfPyspark.datasets import pdbjMineDataset
from urllib.request import urlretrieve
import requests


class PdbjMineSearch(object):
    '''Fetch data using the PDBj Mine 2 SQL service

    Attributes
    ----------
        sqlQuery (str): the sql query [None]
    '''

    URL = "https://pdbj.org/rest/mine2_sql"

    def __init__(self, sqlQuery):

        self.chainLevel = False
        self.pdbIds = []

        dataset = pdbjMineDataset.get_dataset(sqlQuery)

        if dataset == None:
            raise Exception(
                "Dataset empty. Either provide an sql query or a dataset")

        # Check if there is a pdbID file
        if 'structureId' in dataset.columns:
            self.chainLevel = False
            self.pdbIds = [a[0] for a in dataset.select('structureId').collect()]

        if 'structureChainId' in dataset.columns:
            self.chainLevel = True
            ids = [a[0] for a in dataset.select('structureChainId').collect()]
            ids_sub = [i[:4] for i in ids]
            self.pdbIds = ids + ids_sub

    def __call__(self, t):
        match = t[0] in self.pdbIds

        # If results are PDB IDs. but the keys contains chain names,
        # then trucate the chain name before matching (eg. 4HHB.A -> 4HHB)
        if not match and not self.chainLevel and len(t[0]) > 4:
            match = t[0][:4] in self.pdbIds

        return match



#!/user/bin/env python
'''structureToInteractingResidues.py:

Convert a full format of the file to a reduced format

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''
from pyspark.sql import Row


class StructureToInteractingResidues(object):

    def __init__(self, groupName, cutoffDistance):
        self.groupName = groupName
        self.cutoffDistance = cutoffDistance

    def __call__(self, t):
        structureId = t[0]
        structure = t[1]
        groupIndices, groupNames = self._get_group_indices(structure)

        neighbors = []
        for i in range(len(groupName)):

            if groupNames[i] == self.groupName:
                matches = []
                boundingBox = self._calc_bonding_box(
                    structure, groupIndices, i, self.cutoffDistance)
                matches += self._find_neighbors(structure,
                                                index, boudingBox, groupIndices)
                neighbors += self._get_distance_profile(structureId,
                                                        matches, i, groupIndices, groupNames, structure)
        return neighbors

    def _get_distance_profile(self, structureId, matches, index, groupIndices, groupNames, structure):
        cutoffDistanceSq = cutoffDistance * cutoffDistance

        x = structure.x_coord_list
        y = structure.y_coord_list
        z = structure.z_coord_list

        first = groupIndices(index)
        last = groupIndices(index + 1)

        rows = []

        for i in matches:
            if i == index:
                continue

            minDSq = float('inf')
            minIndex = -1

            for j in range(groupIndices[i], groupIndices[i + 1]):

                for k in range(first, last):
                    dx = x[j] - x[k]
                    dy = y[j] - y[k]
                    dz = z[j] - z[k]
                    dSq = dx * dx + dy * dy + dz * dz

                    if (dSq <= cutoffDistanceSq and dSq < minDSq):
                        minDSq = min(minDSq, dSq)
                        minIndex = i

            if minIndex >= 0:
                # TODO add unique group (and atom?) for each group?
                row = Row(structureId,
                          groupNames[index],
                          index,
                          groupNames.get(minIndex),
                          minIndex,
                          float(minDSq**0.5))
                rows.append(row)

        return rows

    def _find_neighbors(self, structure, index, boundingBox, groupIndices):
        x = structure.x_coord_list
        y = structure.y_coord_list
        z = structure.z_coord_list

        matches = []
        for i in range(len(groupIndices) - 1):

            for j in range(groupIndices[i], groupIndices[i + 1]):

                if (x[j] >= boundingBox[0] and x[j] <= boundingBox[1] and
                    y[j] >= boundingBox[2] and y[j] <= boundingBox[3] and
                        z[j] >= boundindBox[4] and z[j] <= boundingBox[5]):
                    matches.append(i)
                    break

        return matches

    def _calc_bonding_box(self, structure, groupIndices, i, cutoffDistance):
        x = structure.x_coord_list
        y = structure.y_coord_list
        z = structure.z_coord_list

        xMin = -float('inf')
        xMax = float('inf')
        yMin = -float('inf')
        yMax = float('inf')
        zMin = -float('inf')
        zMax = float('inf')

        first = groupIndices[i]
        last = groupIndices[i + 1]

        for i in range(first, last):
            xMin = min(xMin, x[i])
            xMax = max(xMax, x[i])
            yMin = min(yMin, y[i])
            yMax = max(yMax, y[i])
            zMin = min(zMin, z[i])
            zMax = max(zMax, z[i])

        boundingBox = [0] * 6
        boundingBox[0] = float(xMin - cutoffDistance)
        boundingBox[1] = float(xMax + cutoffDistance)
        boundingBox[2] = float(yMin - cutoffDistance)
        boundingBox[3] = float(yMax + cutoffDistance)
        boundingBox[4] = float(zMin - cutoffDistance)
        boundingBox[5] = float(zMax + cutoffDistance)

        return boundingBox

    def _get_group_indices(self, structure):
        groupIndices = [0]
        groupNames = []

        atomCounter = 0
        groupCounter = 0
        numChains = structure.chains_per_model[0]

        for i in range(numChains):

            for j in range(structure.groups_per_chain[i]):
                groupIndex = structure.group_type_list[groupCounter]
                groupNames.append(
                    structure.group_list[groupIndex]['groupName'])
                atomCounter += len(
                    structure.group_list[group_Index]['atomNameList'])
                groupIndices.append(atomCounter)
                groupCounter += 1

        return groupIndices, groupNames



#!/usr/bin/env python
'''sparkRegressor.py

Fits a regression model using an MLlib regression method and returns regression
metrics


Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from pyspark.ml import Pipeline
from pyspark.ml.evaluation import RegressionEvaluator
from collections import OrderedDict

class SparkRegressor(object):
    '''Fits a regression model using an MLlib regression method and returns
    regression metrics

    Attributes
    ----------
        predictor: type of multi-class classifier
        label (str): classification label
        testFraction (float): test set fraction [0.3]
        seed (int): random seed
    '''

    def __init__(self, predictor, label, testFraction = 0.3 , seed = 1):

        self.predictor = predictor
        self.label = label
        self.testFraction = testFraction
        self.seed = seed


    def fit(self, data):
        '''Dataset must at least contain the following two columns:
            label : the class labels
            features : feature vector

        Attribute
        ---------
            data (Dataset<Row>)
        Return
        ------
            Dictionary with matrics
        '''

        # Split the data into training and test sets (30% held out for testing)
        splits = data.randomSplit([1.0-self.testFraction, self.testFraction], self.seed)
        trainingData = splits[0]
        testData = splits[1]

        # Train a RandomForest model
        self.predictor.setLabelCol(self.label).setFeaturesCol("features")

        # Chain indexer and forest in a Pipeline
        pipeline = Pipeline().setStages([self.predictor])

        # Train Model. This also runs the indexer.
        model = pipeline.fit(trainingData)

        # Make predictions
        predictions = model.transform(testData)

        # Display some sample predictions
        print(f"Sample predictions: {str(self.predictor).split('_')[0]}") # TODO self.predictor.getClass().getSimpleName
        primaryKey = predictions.columns[0]
        predictions.select(primaryKey, self.label, "prediction").sample(False, 0.1, self.seed).show(50)

        # Collect Metrics
        metrics = OrderedDict()
        metrics["Method"] = str(self.predictor).split("_")[0] # TODO

        evaluator = RegressionEvaluator().setLabelCol(self.label) \
                                         .setPredictionCol("prediction") \
                                         .setMetricName("rmse")
        metrics["rmse"] = str(evaluator.evaluate(predictions))

        return metrics



from .columnarStructure import ColumnarStructure
from .columnarStructureX import ColumnarStructureX
from .mmtfStructure import MmtfStructure
from .dsspSecondaryStructure import DsspSecondaryStructure
from .distanceBox import DistanceBox
from .structureToAllInteractions import StructureToAllInteractions
from .mmtfStructure import MmtfStructure



#!/user/bin/env python
'''g2sDataset.py

This class maps human genetic variation positions to PDB structure positions.
Genomic positions must be specified for the hgvs-grch37 reference genome using
the HGVS sequence variant nomenclature.

References
----------
    HGVS sequence variant nomenclature
        http://varnomen.hgvs.org/
    G2S Web Services
        https://g2s.genomenexus.org/
    Juexin Wang, Robert Sheridan, S Onur Sumer, Nikolaus Schultz, Dong Xu,
    Jianjiong Gao; G2S: a web-service for annotating genomic variants on 3D
    protein structures (2018) Bioinformatics,
        https://doi.org/10.1093/bioinformatics/bty047

Example
-------
    variantIds = ["chr7:g.140449098T>C", "chr7:g.140449100T>C"]
    ds = g2sDataset.get_position_dataset(variantIds, "3TV4", "A")
    ds.show()

    +-----------+-------+-----------+------------+-----------+-------------------+
    |structureId|chainId|pdbPosition|pdbAminoAcid|  refGenome|        variationId|
    +-----------+-------+-----------+------------+-----------+-------------------+
    |       3TV4|      A|        661|           N|hgvs-grch37|chr7:g.140449098T>C|
    |       3TV4|      A|        660|           N|hgvs-grch37|chr7:g.140449100T>C|
    +-----------+-------+-----------+------------+-----------+-------------------+

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from pyspark.sql.functions import col, explode, upper
from pyspark.sql import SparkSession
from pyspark import SparkContext
from io import BytesIO
import requests

REFERENCE_GENOME = "hgvs-grch37"
G2S_REST_URL = "https://g2s.genomenexus.org/api/alignments/"+REFERENCE_GENOME+"/"


def get_position_dataset(variationIds, structureId = None, chainId = None):
    '''Downloads PDB residue mappings for a list of genomic variations

    Attributes
    ----------
        variationIds (list): genomic variation ids, e.g. chr7:g.140449103A>C
        structureId (str): specific PDB structure used for mapping [None]
        chainId (str): specific chain used for mapping [None]

    Returns
    -------
        dataset with PDB mapping information
    '''

    dataset = _get_dataset(variationIds, structureId, chainId)

    if dataset == None: return None

    cols = ["structureId","chainId","pdbPosition","pdbAminoAcid","refGenome",\
            "variationId"]
    return dataset.select(cols).distinct()


def get_full_dataset(variationIds, structureId = None, chainId = None):
    '''Downloads PDB residue mappings and alignment information for a list of
    genomic variations

    Attributes
    ----------
        variationIds (list): genomic variation ids, e.g. chr7:g.140449103A>C
        structureId (str): specific PDB structure used for mapping [None]
        chainId (str): specific chain used for mapping [None]

    Returns
    -------
        dataset with PDB mapping information
    '''

    return _get_dataset(variationIds, structureId, chainId)


def _get_dataset(variationIds, structureId, chainId):
    '''Downloads PDB residue mappings for a list of genomic variations

    Attributes
    ----------
        variationIds (list): genomic variation ids, e.g. chr7:g.140449103A>C
        structureId (str): specific PDB structure used for mapping
        chainId (str): specific chain used for mapping

    Returns
    -------
        dataset with PDB mapping information
    '''

    # Get a spark context
    spark = SparkSession.builder.getOrCreate()
    sc = spark.sparkContext

    # Download data in parallel
    data = sc.parallelize(variationIds) \
             .flatMap(lambda m: _get_data(m, structureId, chainId))

    # Convert Python Rdd to dataframe
    dataframe = spark.read.json(data)

    # Returns None if dataframe is empty
    if len(dataframe.columns) == 0:
        print("g2sDataset: no match found")
        return None

    dataframe = _standardize_data(dataframe)

    return _flatten_dataframe(dataframe)


def _get_data(variationId, structureId, chainId):
    '''Downloads PDB residue mappings for a list of genomic variations
    '''
    data = []

    if structureId is None and chainId is None:
        url = G2S_REST_URL + variationId + '/residueMapping'
    elif structureId is not None and chainId is not None:
        url = G2S_REST_URL + variationId + '/pdb/' + structureId + '_' \
              + chainId + '/residueMapping'
    else:
        raise Exception("Both structureId and chainId have to be provided")

    try:
        req = requests.get(url)
    except:
        print(f"WARNING: could not load data for: {variationId}")
        return data

    if b"\"error\":\"Not Found\"" in req.content:
        print(f"WARNING: could not load data for: {variationId}")
        return data

    results = [inputStream.decode() for inputStream in BytesIO(req.content).readlines()]

    if results is not None and len(results[0]) > 100:
        results = _add_variant_id(results, REFERENCE_GENOME, variationId)
        data = results

    return data


def _add_variant_id(json, refGenome, variationId):
    '''Adds reference genome and variationId to each json records

    Attributes
    ----------
        json: list of original json strings
        refGenome: reference genome
        variationId: variation identifier
    '''

    ids = "\"refGenome\":\"" + refGenome + "\"," + "\"variationId\":\"" \
          + variationId + "\"," + "\"alignmentId\""
    return [j.replace("\"alignmentId\"", ids) for j in json]


def _standardize_data(df):
    '''Standardize data and column names to be consistent
    '''
    return df.withColumn("structureId", upper(col("pdbId"))) \
             .withColumnRenamed("chain", "chainId")


def _flatten_dataframe(df):
    return df.withColumn("pdbPosition", explode(col("residueMapping.pdbPosition"))) \
             .withColumn("pdbAminoAcid", explode(col("residueMapping.pdbAminoAcid")))



#!/user/bin/env python
'''dsspSecondaryStructure.py:

Authorship information:
    __author__ = "Yue Yu"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from enum import Enum


class DsspSecondaryStructure(Enum):

    PI_HELIX = 0
    BEND = 1
    ALPHA_HELIX = 2
    EXTENDED = 3
    THREE_TEN_HELIX = 4
    BRIDGE = 5
    TURN = 6
    COIL = 7

    def get_q3_code(numericCode):

        cases = {0: DsspSecondaryStructure.ALPHA_HELIX,
                 1: DsspSecondaryStructure.COIL,
                 2: DsspSecondaryStructure.ALPHA_HELIX,
                 3: DsspSecondaryStructure.EXTENDED,
                 4: DsspSecondaryStructure.ALPHA_HELIX,
                 5: DsspSecondaryStructure.EXTENDED,
                 6: DsspSecondaryStructure.COIL,
                 7: DsspSecondaryStructure.COIL
                 }
        if numericCode in cases:
            return cases[numericCode]
        else:
            return DsspSecondaryStructure.COIL

    def get_one_letter_code(self):
        cases = {0: '5',
                 1: 'S',
                 2: 'H',
                 3: 'E',
                 4: 'G',
                 5: 'B',
                 6: 'T',
                 7: 'C'
                 }
        return cases[self.value]

    def get_dssp_code(numericCode):
        for x in list(DsspSecondaryStructure):
            if x.value == numericCode:
                return x
        return DsspSecondaryStructure.COIL



'''mmtfDecoder.py

Provides efficient methods to decode mmtf structures

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''

import numpy as np


def run_length_decoder_numpy(in_array):
    """Decodes a run length encoded array

    Attributes
    ----------
        in_array (list): the input list to apply run length decoder on
    """

    lengths = np.array(in_array[1::2])
    values = np.array(in_array[0::2])
    starts = np.insert(np.array([0]), 1, np.cumsum(lengths))[:-1]
    ends = starts + lengths
    n = ends[-1]
    x = np.full(n, np.nan)
    for l, h, v in zip(starts, ends, values):
        x[l:h] = v
    return x


def recursive_index_decode(int_array, decode_num=1000):
    """Unpack an array of integers using recursive indexing.

    Attribute
    ---------
        int_array (list): the input array of integers
        decode_num (int): the number used for decoding [1000]

    Returns
    -------
        return the array of integers after recursive index decoding
    """

    maximum = 32767
    minimum = -32768
    out_arr = np.cumsum(int_array) / decode_num
    return out_arr[(int_array != maximum) & (int_array != minimum)]


def decode_entity_list(input_data):
    """Convert byte strings to strings in the entity list.

    Attributes
    ----------
        input_data (list): the list of entities

    Returns
    -------
        return the decoded entity list
    """

    out_data = []
    for entry in input_data:
        out_data.append(convert_entity(entry))
    return out_data


def decode_group_list(input_data):
    """Convert byte strings to strings in the group map.

    Attributes
    ----------
        input_data (list): the list of groups

    Returns
    -------
        return the decoded group list
    """

    out_data = []
    for entry in input_data:
        out_data.append(convert_group(entry))
    return out_data


def convert_group(input_group):
    """Convert an individual group from byte strings to regula strings.

    Attributes
    ----------
        input_group (list): the list of input groups

    Returns
    -------
        return the decoded group
    """

    output_group = {}
    for key in input_group:
        if key in [b'elementList', b'atomNameList']:
            output_group[key.decode('ascii')] = [x.decode('ascii')
                                                 for x in input_group[key]]
        elif key in [b'chemCompType', b'groupName', b'singleLetterCode']:
            output_group[key.decode(
                'ascii')] = input_group[key].decode('ascii')
        else:
            output_group[key.decode('ascii')] = input_group[key]
    return output_group


def convert_entity(input_entity):
    """Convert an individual entity from byte strings to regular strings

    Attributes
    ----------
        input_entity (list): the list of entities to decode

    Returns
    -------
        return the decoded entity
    """

    output_entity = {}
    for key in input_entity:
        if key in [b'description', b'type', b'sequence']:
            output_entity[key.decode(
                'ascii')] = input_entity[key].decode('ascii')
        else:
            output_entity[key.decode('ascii')] = input_entity[key]
    return output_entity



#!/user/bin/env python
'''pdbjMineDataset.py

This filter runs an PDBj Mine 2 Search web service using SQL query

References
----------
    Data are provided through Mine2 SQL
        https://pdbj.org/help/mine2-sql
    Queries can be designed with the interactive PDBj Mine 2 query service:
        https://pdbj.org/mine/sql
    PDB metadata are described in the PDB mmCIF Dictionary:
        http://mmcif.wwpdb.org/

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

import urllib
import tempfile
from pyspark.sql import SparkSession
from pyspark.sql.functions import col, lit, upper, concat
from urllib.request import urlretrieve
import requests


def get_dataset(sqlQuery):
    '''Runs a PDBj Mine 2 search web service using an SQL query

    Attributes
    ----------
        sqlQuery (str): the sql query for the web service
    '''

    encodedSQL = urllib.parse.quote(sqlQuery)
    tmp = tempfile.NamedTemporaryFile(delete=False)

    URL = "https://pdbj.org/rest/mine2_sql"
    urlretrieve(URL + "?format=csv&q=" + encodedSQL, tmp.name)

    spark = SparkSession.builder.getOrCreate()

    ds = spark.read.format("csv") \
        .option("header", "true") \
        .option("inferSchema", "true") \
        .option("parserLib", "UNIVOCITY") \
        .load(tmp.name)

    # rename.concatenate columns to assign
    # consistent primary keys to datasets
    if "pdbid" in ds.columns:
        # this project uses upper case pdbids
        ds = ds.withColumn("pdbid", upper(col("pdbid")))

        if "chain" in ds.columns:
            ds = ds.withColumn("structureChainId", \
                               concat(col("pdbid"), lit("."), col("chain")))
            ds.drop("pdbid","chain")
        else:
            ds = ds.withColumnRenamed("pdbid", "structureId")

    return ds



#!/user/bin/env python
'''containsDSaccharideChain.py

This filter returns entries that contain chain(s) made of linear and branched D-saccharides.
The default constructor returns entries that contain at least one
polymer chain that is a D-saccharides. If the "exclusive" flag is set to true
in the constructor, all polymer chains must be D-saccharides. For a multi-model structure,
this filter only checks the first model.

Note: Since the PDB released PDBx/mmCIF version 5.0 in July 2017, it appears that
all polysaccharides have been converted to monomers. Therefore, this filter
does not return any results.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from mmtfPyspark.filters import ContainsPolymerChainType


class ContainsDSaccharideChain(object):
    '''Default constructor matches any entry that contains at least one chain
    made of D-saccharides. As an example, a glycosylated protein complex passes
    this filter.

    Optional constructor that can be used to filter entries that exclusively
    contain D-saccharide chains. For example, with "exclusive" set to true, an
    D-saccharide/protein does not pass this filter.

    Attributes
    ----------
        exclusive (bool): if true, only return entries that are exclusively
                          contain D-saccharide chains

    '''

    def __init__(self, exclusive=False):
        self.filter = ContainsPolymerChainType([
            ContainsPolymerChainType.D_SACCHARIDE,
            ContainsPolymerChainType.SACCHARIDE,
            ContainsPolymerChainType.D_SACCHARIDE_14_and_14_LINKING,
            ContainsPolymerChainType.D_SACCHARIDE_14_and_16_LINKING
        ], exclusive=exclusive)

    def __call__(self, t):
        return self.filter(t)



#!/user/bin/env python
'''ligandInteractionFIngerprint.py

Finds interactions between ligands and polymer chains and maps them onto polymer sequences.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''

from mmtfPyspark.interactions import InteractionFilter, AtomInteraction, InteractionCenter
from mmtfPyspark.utils import ColumnarStructure
from mmtfPyspark.utils import DistanceBox
from pyspark.sql import Row
import numpy as np


class LigandInteractionFingerprint(object):

    def __init__(self, interactionFilter):
        self.filter = interactionFilter

    def __call__(self, t):

        structureId = t[0]
        structure = t[1]
        return self.get_interactions(structureId, structure)

    def get_interactions(self, structureId, structure):
        rows = []

        cutoffDistanceSquared = self.filter.get_distance_cutoff() ** 2
        arrays = ColumnarStructure(structure, True)

        chainNames = arrays.get_chain_names()
        groupNames = arrays.get_group_names()
        groupNumbers = arrays.get_group_numbers()
        atomNames = arrays.get_atom_names()
        entityIndices = arrays.get_entity_indices()
        elements = arrays.get_elements()
        polymer = arrays.is_polymer()

        sequenceMapIndices = arrays.get_sequence_positions()
        x = arrays.get_x_coords()
        y = arrays.get_y_coords()
        z = arrays.get_z_coords()

        # create a distance box for quick lookup interactions of polymer atoms
        # of the specified elements
        box = DistanceBox(self.filter.get_distance_cutoff())
        for i in range(arrays.get_num_atoms()):

            if polymer[i] \
                and self.filter.is_target_group(groupNames[i]) \
                and self.filter.is_target_atom_name(atomNames[i]) \
                and self.filter.is_target_element(elements[i]) \
                and not self.filter.is_prohibited_target_group(groupNames[i]):

                newPoint = np.array([x[i],y[i],z[i]])
                box.add_point(newPoint, i)

        groupToAtomIndices = arrays.get_group_to_atom_indices()

        for g in range(arrays.get_num_groups()):

            # position of first and last atom +1 in group
            start = groupToAtomIndices[g]
            end = groupToAtomIndices[g+1]

            # skip polymer groups
            if polymer[start]:
                continue

            # the specified filter conditions (some groups may be excluded,
            # e.g. water)
            if self.filter.is_query_group(groupNames[start]):

                print(groupNames[start])
                # create list of atoms that interact within the cutoff distance
                neighbors = []
                for a in range(start,end):

                    if self.filter.is_query_atom_name(atomNames[a]) \
                        and self.filter.is_query_element(elements[a]):

                        p = np.array([x[a], y[a], z[a]])

                        # loop up neighbors that are within a cubic
                        for j in box.get_neighbors(p):
                            dx = x[j] - x[a]
                            dy = y[j] - y[a]
                            dz = z[j] - z[a]
                            dSq = dx * dx + dy * dy + dz * dz

                            if dSq <= cutoffDistanceSquared:
                                neighbors.append(j)

                if len(neighbors) == 0:
                    continue

                interactions2 = {}
                for neighbor in neighbors:

                    if chainNames[neighbor] not in interactions2:
                        interactions2[chainNames[neighbor]] = []

                    # keep track of which group is interacting
                    seqPos = sequenceMapIndices[neighbor]

                    # non-polymer groups have a negative index and are exlcuded here
                    if seqPos > 0:
                        l = [seqPos, groupNumbers[neighbor], entityIndices[neighbor]]
                        interactions2[chainNames[neighbor]].append(l)

                for key, val in interactions2.items():

                    sequenceIndices = set()
                    residueNames = set()
                    sequence = None

                    for v in val:
                        sequenceIndices.add(int(v[0]))
                        residueNames.add(int(v[1]))
                        if sequence is None:
                            sequence = structure.entity_list[v[2]]['sequence']

                    if len(sequenceIndices) > 0:
                        rows.append(Row(structureId + "." + key, groupNames[start], \
                                        groupNumbers[start], chainNames[start], \
                                        key, sorted(list(residueNames)), \
                                        sorted(list(sequenceIndices)), sequence,\
                                        len(interactions2)))
        return rows



#!/user/bin/env python
'''structureToPolymerChain.py:

Maps a structure to its individual polymer chains. Polymer chains
include polypeptides, polynucleotides, and linear and branched polysaccharides.
For a multi-model structure, only the first model is considered.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "debug"
'''
from mmtf.utils import *
from mmtf.api.mmtf_writer import MMTFEncoder
from numpy import *


class StructureToBioassembly(object):
    '''Extracts all polymer chains from a structure. If the argument is set to true,
    the assigned key is: <PDB ID.Chain ID>, where Chain ID is the unique identifier
    assigned to each molecular entity in an mmCIF file. This Chain ID corresponds to
    <a href="http://mmcif.wwpdb.org/dictionaries/mmcif_mdb.dic/Items/_atom_site.label_asym_id.html">
    _atom_site.label_asym_id</a> field in an mmCIF file.

    Attributes
    ----------
        useChainIdInsteadOfChainName (bool): if true, use Chain Id in the key assignments
        excludeDuplicates (bool): if true return only one chain for each sequence
    '''

    def __init__(self, useChainIdInsteadOfChainName=False, excludeDuplicates=False):
        self.useChainIdInsteadOfChainName = useChainIdInsteadOfChainName
        self.excludeDuplicates = excludeDuplicates

    def __call__(self, t):
        structure = t[1]
        if not structure.alt_loc_set:
            structure = structure.set_alt_loc_list()
        # print(structure.group_type_list)

        bioassemblies = structure.bio_assembly

        numBioassembly = len(bioassemblies)

        resList = list()

        for i in range(numBioassembly):
            bioAssembly = MMTFEncoder()
            structureId = structure.structure_id + '-BioAssembly' + \
                bioassemblies[i]['name']
            totAtoms = 0
            totBonds = 0
            totGroups = 0
            totChains = 0
            totModels = structure.num_models
            numTrans = len(bioassemblies[i]['transformList'])
            bioChainList = [[]] * numTrans
            transMatrix = [[]] * numTrans
            for ii in range(numTrans):

                bioChainList[ii] = bioassemblies[i]['transformList'][ii]['chainIndexList']
                transMatrix[ii] = bioassemblies[i]['transformList'][ii]['matrix']

                for j in range(totModels):
                    totChains = totChains + len(bioChainList[ii])
                    groupCounter = 0
                    for k in range(structure.chains_per_model[j]):

                        adding = False
                        for currChain in bioChainList[ii]:
                            if currChain == k:
                                adding = True
                        if adding:
                            totGroups = totGroups + \
                                structure.groups_per_chain[k]
                        for h in range(structure.groups_per_chain[k]):
                            if adding:
                                groupIndex = structure.group_type_list[groupCounter]
                                totAtoms = totAtoms + \
                                    len(structure.group_list[groupIndex]
                                        ['atomNameList'])
                                totBonds = totBonds + \
                                    len(structure.group_list[groupIndex]
                                        ['bondOrderList'])
                            groupCounter = groupCounter + 1
            # Set header
            bioAssembly.init_structure(
                totBonds, totAtoms, totGroups, totChains, totModels, structureId)
            decoder_utils.add_xtalographic_info(structure, bioAssembly)
            decoder_utils.add_header_info(structure, bioAssembly)

            modelIndex = 0
            chainIndex = 0
            groupIndex = 0
            atomIndex = 0
            chainCounter = 0
            description = {}

            for ii in range(totModels):
                numChainsPerModel = structure.chains_per_model[modelIndex] * numTrans
                bioAssembly.set_model_info(modelIndex, numChainsPerModel)
                chainToEntityIndex = self._getChainToEntityIndex(structure)

                for j in range(structure.chains_per_model[modelIndex]):
                    currGroupIndex = groupIndex
                    currAtomIndex = atomIndex
                    for k in range(numTrans):
                        currChainList = bioChainList[k]
                        currMatrix = transMatrix[k]
                        addThisChain = False
                        for currChain in currChainList:
                            if currChain == j:
                                addThisChain = True
                        groupIndex = currGroupIndex
                        atomIndex = currAtomIndex
                        xCoords = structure.x_coord_list
                        yCoords = structure.y_coord_list
                        zCoords = structure.z_coord_list

                        m = reshape(matrix(currMatrix), (4, 4))

                        if addThisChain:
                            entityToChainIndex = chainToEntityIndex[chainIndex]
                            if structure.entity_list[entityToChainIndex]['description'] in description:
                                index = description[structure.entity_list[entityToChainIndex]['description']]
                                bioAssembly.entity_list[index]['chainIndexList'].append(chainCounter)
                            else:
                                bioAssembly.set_entity_info([chainCounter],
                                                            structure.entity_list[entityToChainIndex]['sequence'],
                                                            structure.entity_list[entityToChainIndex]['description'],
                                                            structure.entity_list[entityToChainIndex]['type'])
                                description[bioAssembly.entity_list[-1]['description']] = len(bioAssembly.entity_list)-1
                            bioAssembly.set_chain_info(structure.chain_id_list[chainIndex],
                                                       structure.chain_name_list[chainIndex],
                                                       structure.groups_per_chain[chainIndex])
                            chainCounter = chainCounter + 1
                        for jj in range(structure.groups_per_chain[chainIndex]):
                            # print(structure.group_type_list)
                            currgroup = structure.group_type_list[groupIndex]
                            # if ii == 0 and j == 0 and jj < 10:
                            # print(currgroup)

                            if addThisChain:
                                bioAssembly.set_group_info(structure.group_list[currgroup]['groupName'],
                                                           structure.group_id_list[groupIndex],
                                                           structure.ins_code_list[groupIndex],
                                                           structure.group_list[currgroup]['chemCompType'],
                                                           len(
                                                               structure.group_list[currgroup]['atomNameList']),
                                                           len(
                                                               structure.group_list[currgroup]['bondOrderList']),
                                                           structure.group_list[currgroup]['singleLetterCode'],
                                                           structure.sequence_index_list[groupIndex],
                                                           structure.sec_struct_list[groupIndex])
                            for kk in range(len(structure.group_list[currgroup]['atomNameList'])):
                                if addThisChain:
                                    p1 = array(
                                        [xCoords[atomIndex], yCoords[atomIndex], zCoords[atomIndex], 1])
                                    p2 = matmul(p1, m)
                                    bioAssembly.set_atom_info(
                                        structure.group_list[currgroup]['atomNameList'][kk],
                                        structure.atom_id_list[atomIndex],
                                        structure.alt_loc_list[atomIndex],
                                        p2.item(0),
                                        p2.item(1),
                                        p2.item(2),
                                        structure.occupancy_list[atomIndex],
                                        structure.b_factor_list[atomIndex],
                                        structure.group_list[currgroup]['elementList'][kk],
                                        structure.group_list[currgroup]['formalChargeList'][kk], )

                                atomIndex = atomIndex + 1

                            # bond not implemented
                            if addThisChain:

                                for l in range(len(structure.group_list[currgroup]['bondOrderList'])):

                                    bondIndOne = structure.group_list[currgroup]['bondAtomList'][l * 2]
                                    bondIndTwo = structure.group_list[currgroup]['bondAtomList'][l * 2 + 1]
                                    bondOrder = structure.group_list[currgroup]['bondOrderList'][l]

                                    #newChain.set_group_bond(bondIndOne, bondIndTwo, bondOrder)

                                    bioAssembly.current_group.bond_atom_list += [
                                        bondIndOne, bondIndTwo]
                                    bioAssembly.current_group.bond_order_list.append(
                                        bondOrder)

                            groupIndex = groupIndex + 1

                    chainIndex = chainIndex + 1
                modelIndex = modelIndex + 1
                # print(type(currMatrix))
            bioAssembly.finalize_structure()
            resList.append((structureId, bioAssembly))
        return resList

    def _getNumAtomsAndBonds(self, structure):
        '''Gets the number of atoms and bonds per chain
        '''
        numChains = structure.chains_per_model[0]
        atomsPerChain = [0] * numChains
        bondsPerChain = [0] * numChains
        groupCounter = 0

        for i in range(numChains):

            for j in range(structure.groups_per_chain[i]):
                groupIndex = structure.group_type_list[groupCounter]
                atomsPerChain[i] = len(
                    structure.group_list[groupIndex]['atomNameList'])
                bondsPerChain[i] = len(
                    structure.group_list[groupIndex]['bondOrderList'])
                groupCounter += 1

        return atomsPerChain, bondsPerChain

    def _getChainToEntityIndex(self, structure):
        '''Returns an list that maps a chain index to an entity index.

        Attributes
        ----------
            structure: structureDataInterFace
        '''
        entityChainIndex = [0] * structure.num_chains

        for i in range(len(structure.entity_list)):

            for j in structure.entity_list[i]["chainIndexList"]:
                entityChainIndex[j] = i

        return entityChainIndex



#!/user/bin/env python
'''containsGroup.py:

This filter returns entries that contain at least one of the specified groups
(residues).Groups are specified by their one, two, or three-letter codes,
e.g. "F", "MG", "ATP", as defined in the PDB Chemical Component Dictionary.

References
----------
    PDB Chemical Component Dictionary:
        https://www.wwpdb.org/data/ccd

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''


class ContainsGroup(object):
    '''Returns entries that contain at least one of the specified groups

    Attributes
    ----------
        groupQuery (list[str]): list of group names
    '''

    def __init__(self, *args):
        groups = [a for a in args]
        self.groupQuery = set(groups)

    def __call__(self, t):
        groups = [t[1].group_list[idx]['groupName']
                  for idx in t[1].group_type_list]

        for group in self.groupQuery:
            if group in groups:
                return True

        return False



#!/usr/bin/env python
'''datasetClassifier.py

Runs binary and multi-class classifiers on a given dataset.
Dataset are read as Parquet file. The dataset must contain
a feature vector named "features" and a classification column.
The column name of the classification column must be specified
on the command lines.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from mmtfPyspark.ml import SparkMultiClassClassifier, datasetBalancer
from pyspark.sql import SparkSession
from pyspark.ml.classification import DecisionTreeClassifier, LogisticRegression, MultilayerPerceptronClassifier, RandomForestClassifier
import sys
import time

def main(argv):

    # Name of prediction column
    label = argv[1]

    start = time.time()

    spark = SparkSession.builder \
                        .master("local[*]") \
                        .appName("datasetClassifier") \
                        .getOrCreate()

    data = spark.read.parquet(argv[0]).cache()

    vector = data.first()["features"]
    featureCount = len(vector)

    print(f"Feature count    : {featureCount}")
    classCount = int(data.select(label).distinct().count())
    print(f"Class count    : {classCount}")
    print(f"Dataset size (unbalanced)    : {data.count()}")
    data.groupby(label).count().show(classCount)

    data = datasetBalancer.downsample(data, label, 1)

    print(f"Dataset size (balanced)  : {data.count()}")
    data.groupby(label).count().show(classCount)

    testFraction = 0.3
    seed = 123

    # DecisionTree
    dtc = DecisionTreeClassifier()
    mcc = SparkMultiClassClassifier(dtc, label, testFraction, seed)
    matrics = mcc.fit(data)
    for k,v in matrics.items(): print(f"{k}\t{v}")

    # RandomForest
    rfc = RandomForestClassifier()
    mcc = SparkMultiClassClassifier(rfc, label, testFraction, seed)
    matrics = mcc.fit(data)
    for k,v in matrics.items(): print(f"{k}\t{v}")

    # LogisticRegression
    lr = LogisticRegression()
    mcc = SparkMultiClassClassifier(lr, label, testFraction, seed)
    matrics = mcc.fit(data)
    for k,v in matrics.items(): print(f"{k}\t{v}")

    # MultilayerPerceptronClassifier
    layers = [featureCount, 10, classCount]
    mpc = MultilayerPerceptronClassifier().setLayers(layers) \
                                          .setBlockSize(128) \
                                          .setSeed(1234) \
                                          .setMaxIter(200)
    mcc = SparkMultiClassClassifier(mpc, label, testFraction, seed)
    matrics = mcc.fit(data)
    for k,v in matrics.items(): print(f"{k}\t{v}")

    end = time.time()
    print("Time: %f  sec." %(end-start))


if __name__ == "__main__":

    if len(sys.argv) < 3:
        raise Exception("python datasetClassifier.py <parquet file> <prediction column name>")
        sys.exit()

    main(sys.argv[1:])



#!/user/bin/env python
'''structureToSecondaryStructureSegments.py:

Maps chain seuqnce to its sequence segments

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''
from pyspark.sql import Row


class StructureToSecondaryStructureSegments(object):
    '''Constructor sets the segment length.

    Attributes
    ----------
        length (int): length of the secondary structure segments
    '''

    def __init__(self, length):
        if length % 2 != 1:
            print("length has to be an odd number")

        self.length = length

    def __call__(self, t):
        structureChainId = t[0]
        sequence = t[1]
        dsspQ3 = t[6]
        dsspQ8 = t[5]

        numSegments = max(0, len(sequence) - self.length)
        sequences = []

        for i in range(len(sequence) - self.length):
            currSeq = sequence[i:i + self.length]

            # print(dsspQ3)
            labelQ3 = dsspQ3[i + int(self.length / 2): i + int(self.length / 2) + 1]
            labelQ8 = dsspQ8[i + int(self.length / 2): i + int(self.length / 2) + 1]

            if (labelQ8 != "X" and labelQ3 != "X"):
                sequences.append(
                    Row(structureChainId, currSeq, labelQ8, labelQ3))

        return sequences



#!/user/bin/env python
'''myVariantDataset.py

This class queries and retrieves missense variations using the MyVariant.info
web services for a list of UniProt ids.

References
----------
    For more information:
        http://myvariant.info
    Query syntax:
        http://myvariant.info/docs/
    Xin J, Mark A, Afrasiabi C, Tsueng G, Juchler M, Gopal N, Stupp GS, Putman
    TE, Ainscough BJ, Griffith OL, Torkamani A, Whetzel PL, Mungall CJ, Mooney
    SD, Su AI, Wu C (2016) High-performance web services for querying gene and
    variant annotation. Genome Biology 17(1):1-7
        https://doi.org/10.1186/s13059-016-0953-9


Example
-------
    Get all missense variations for a list of Uniprot Ids:
        uniprotIds = ['P15056']    # BRAF
        ds = MyVariantDataset.get_variations(uniprotIds)
        ds.show()

    Return missense variations that match a query
        uniprotIds = ['P15056']    # BRAF
        query = "clinivar.rcv.clinical_significance:pathogenic" \
                + "OR linivar.rcv.clinical_significance:likely pathogenic"
        ds = MyVariantDataset.get_variations(uniprotIds, query)
        ds.show()

        +-------------------+---------+
        |        variationId|uniprotId|
        +-------------------+---------+
        |chr7:g.140454006G>T|   P15056|
        |chr7:g.140453153A>T|   P15056|
        |chr7:g.140477853C>A|   P15056|


Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from pyspark.sql import SparkSession
from pyspark.sql.functions import col, explode
from pyspark import SparkContext
from io import BytesIO
import requests


MYVARIANT_QUERY_URL = "http://myvariant.info/v1/query?q="
MYVARIANT_SCROLL_URL = "http://myvariant.info/v1/query?scroll_id="


def get_variations(uniprotIds, query = ''):
    '''Returns a dataset of missense variabtions for a list of Uniprot Ids and a
    MyVariant.info query.

    References
    ----------
        query syntax
            http://myvariant.info/docs/

    Example
    -------
        uniprotIds = ['P15056']    # BRAF
        query = "clinivar.rcv.clinical_significance:pathogenic" \
                + "OR linivar.rcv.clinical_significance:likely pathogenic"
        ds = MyVariantDataset.get_variations(uniprotIds, query)

    Attributes
    ----------
        uniprotIds (list): list of Uniprot Ids
        query (str): MyVariant.info query string ['']

    Returns
    -------
        dataset with variation Ids and Uniprot Ids or null if no data are found
    '''

    # Get spark context
    spark = SparkSession.builder.getOrCreate()
    sc = spark.sparkContext

    # Download data in paralle
    data = sc.parallelize(uniprotIds).flatMap(lambda x: _get_data(x, query))

    # Convert Python Rdd to dataframe
    dataframe = spark.read.json(data)

    # return null if dataset contains no results
    if 'hits' not in dataframe.columns:
        print("MyVariantDataset: no match found")
        return None

    return _flatten_dataframe(dataframe)


def _get_data(uniprotId, query):
    '''get data from MyVariant
    '''

    data = []

    # Create url
    url = MYVARIANT_QUERY_URL \
          + "snpeff.ann.effect:missense_variant AND dbnsfp.uniprot.acc:" \
          + uniprotId
    if query is not '':
        url += f" AND ({query})"
    url += "&fields=_id&fetch_all=true"
    url = url.replace(" ","%20")

    # Open url
    try:
        req = requests.get(url)
    except:
        print(f"WARNING: Counld not load data for {uniprotId}")
        return data
    inputStream = BytesIO(req.content)

    # read results
    # A maximum of 1000 records are returned per request
    # Additional records are retrived iteratively using the scrollId
    results = _read_results(inputStream)

    # check if hits results are empty
    if "[]" not in results:
        results = _add_uniprot_id(results, uniprotId)
        data.append(results)

        scrollId = _get_scroll_id(results)

        while scrollId is not None:
            url = MYVARIANT_SCROLL_URL + scrollId
            try:
                req = requests.get(url)
            except:
                print(f"WARNING: Counld not load data for {uniprotId}")
                continue

            inputStream = BytesIO(req.content)
            results = _read_results(inputStream)

            if "No results to return" not in results:
                results = _add_uniprot_id(results, uniprotId)
                data.append(results)
                scollId = _get_scroll_id(results)
            else:
                scrollId = None

    return data


def _get_scroll_id(results):
    '''Get the scroll id from results
    '''
    if "_scroll_id" in results:
        return results.split("\"")[3]
    return None


def _read_results(inputStream):
    '''Converts data read from input stream into a single line of text

    Attributes
    ----------
        inputStream (is): input stream

    Returns
    -------
        response as a string
    '''

    sb = ''
    for line in inputStream:
        sb += line.decode()
    return sb


def _add_uniprot_id(line, uniprotId):
    ids = "\"uniprotId\":" + "\"" + uniprotId + "\"," + "\"hits\"";
    return line.replace("\"hits\"", ids);
    #ids = f"\"uniprotIds\":\"{uniprotId}\",\n      \"_id\""
    #return line.replace("\"_id\"", ids)


def _flatten_dataframe(df):
    return df.withColumn("variationId", explode(df.hits._id)) \
             .select(col("variationId"), col("uniprotId"))



#!/user/bin/env python
'''depositionDate.py

This filter return true if the deposition date of this structure is within the
specified range

authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from dateutil.parser import parse


class DepositionDate(object):

    def __init__(self, startdate, enddate):
        '''This filter return True if the deposition date of this structure is
        within the specified range

        Attributes
        ----------
            startdate (str): start of the deposition date range
            enddate (str): end of the deposition date range
        '''
        self.startdate = parse(startdate)
        self.enddate = parse(enddate)

    def __call__(self, t):
        structure = t[1]
        depositiondate = parse(structure.deposition_date)

        return depositiondate >= self.startdate and depositiondate <= self.enddate



#!/user/bin/env python
'''notFilter.py

This filter wraps another filter and negates its result

authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''


class NotFilter(object):
    '''Constructor takes another filter as input

    Attributes
    ----------
        filter1 (filter): first filter to be negated
        filter2 (filter): second filter to be negated
    '''

    def __init__(self, filter_function):
        self.filter = filter_function

    def __call__(self, t):
        return not self.filter(t)



#!/user/bin/env python
'''viewStructures.py

Simple wrapper functions that uses ipywidgets and py3Dmol to view a list of
protein structures.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from ipywidgets import interact, IntSlider, Dropdown
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import py3Dmol


def view_structure(pdbIds, bioAssembly = False, style='cartoon', color='spectrum'):
    '''A wrapper function that simply displays a list of protein structures using
    ipywidgets and py3Dmol

    Attributes
    ----------
        pdbIds (list<Strings>) : A list of PDBIDs to display
        bioAssembly (Bool): display bioAssembly
        style : Style of 3D structure (stick line cross sphere cartoon VDW MS)
        color : Color of 3D structure
    '''

    if type(pdbIds) == str:
        pdbIds = [pdbIds]

    def view3d(i=0):
        '''Simple structure viewer that uses py3Dmol to view PDB structure by
        indexing the list of PDBids

        Attributes
        ----------
            i (int): index of the protein if a list of PDBids
        '''
        print(f"PdbID: {pdbIds[i]}, Style: {style}")


        if '.' not in pdbIds[i]:
            viewer = py3Dmol.view(query='pdb:' + pdbIds[i], options={'doAssembly': bioAssembly})
            viewer.setStyle({style: {'color': color}})
            viewer.setStyle({'hetflag': True},{'stick':{'singleBond':False}})

        else:
            pdbid,chainid = pdbIds[i].split('.')
            viewer = py3Dmol.view(query='pdb:' + pdbid, options={'doAssembly': bioAssembly})
            viewer.setStyle({})
            viewer.setStyle({'chain': chainid}, {style: {'color': color}})
            viewer.setStyle({'chain': chainid, 'hetflag': True},{'stick':{'singleBond':False}})
            viewer.zoomTo({'chain': chainid})

        return viewer.show()

    s_widget = IntSlider(min=0, max=len(pdbIds)-1, description='Structure', continuous_update=False)
    return interact(view3d, i=s_widget)


def view_group_interaction(pdbIds, interacting_group='None', style='cartoon', color='spectrum'):
    '''A wrapper function that simply displays a list of protein structures using
    ipywidgets and py3Dmol and highlight specified interacting groups

    Attributes
    ----------
        pdbIds (list<Strings>) : A list of PDBIDs to display
        interacting_atom (String) : The interacting atom to highlight
        style : Style of 3D structure (stick line cross sphere cartoon VDW MS)
        color : Color of 3D structure
    '''

    if type(pdbIds) == str:
        pdbIds = [pdbIds]

    def view3d(i=0):
        '''Simple structure viewer that uses py3Dmol to view PDB structure by
        indexing the list of PDBids

        Attributes
        ----------
            i (int): index of the protein if a list of PDBids
        '''

        print(
            f"PdbID: {pdbIds[i]}, Interactions: {interacting_group}, Style: {style}")

        viewer = py3Dmol.view(query='pdb:' + pdbIds[i])
        viewer.setStyle({style: {'color': color}})

        if interacting_group != "None":

            viewer.setStyle({'resn': interacting_group}, {
                            'sphere': {}})

        return viewer.show()

    s_widget = IntSlider(min=0, max=len(pdbIds)-1, description='Structure', continuous_update=False)
    return interact(view3d, i=s_widget)


def view_binding_site(pdbIds=None, groups=None, chains=None, distance=3.0):
    '''A wrapper function that zooms in to a group of a protein structure and highlight
    its neighbors within a certain distance.

    Attributes
    ----------
        pdbIds (list<Strings>, String) : A list of PDBIDs to display
        groups (list<int>) : A list of groups to center at for each protein structure
        chains (list<char>) : A list of chains specified for each protein structure.
                              If no chains is specified, chain 'A' will be default to
                              all structures.
        cutoffDistance (float) : The cutoff distance use the find the neighbors
                                 of specified group
    '''

    if pdbIds is None or groups is None:
        raise ValueError("PdbIds and groups need to be specified")

    if len(pdbIds) != len(groups):
        raise ValueError(
            "Number of structures should match with number of groups")

    if type(pdbIds) == str and groups == str:
        pdbIds, groups = [pdbIds], [groups]

    if chains is None:
        chains = ['A'] * len(pdbIds)

    def view3d(i=0):
        '''Simple structure viewer that zooms into a specified group and highlight
        its neighbors

        Attributes
        ----------
            i (int): index of the protein if a list of PDBids
        '''

        print(
            f"PDB: {pdbIds[i]}, group: {groups[i]}, chain: {chains[i]}, cutoffDistance: {distance}")

        if type(groups[i]) == int:
            center = {'resi': groups[i], 'chain': chains[i]}
            neighbors = {'resi': groups[i], 'chain': chains[i],
                         'byres': 'true', 'expand': distance}
        else:
            center = {'resn': groups[i], 'chain': chains[i]}
            neighbors = {'resn': groups[i], 'chain': chains[i],
                         'byres': 'true', 'expand': distance}


        viewer = py3Dmol.view(query='pdb:' + pdbIds[i])
        viewer.setStyle(neighbors, {'stick': {}});
        viewer.setStyle(center, {'sphere': {'colorscheme': 'orangeCarbon'}})
        viewer.zoomTo(neighbors)

        return viewer.show()

    s_widget = IntSlider(min=0, max=len(pdbIds)-1, description='Structure', continuous_update=False)
    return interact(view3d, i=s_widget)


def group_interaction_viewer(df, sortBy, metal = None):
    '''A wrapper function that zooms in to a group in a protein structure and
    highlight its interacting atoms. The input dataframe should be generated
    from the GroupInteractionExtractor class.

    References
    ----------
        GroupInteractionExtractor:
        https://github.com/sbl-sdsc/mmtf-pyspark/blob/master/mmtfPyspark/interactions/groupInteractionExtractor.py

    Attributes
    ----------
        df (dataframe): the dataframe generated from GroupIneteractionExtractor
        sort_by (str): the q value to sort by ['q4','q5','q6']
    '''

    # Filter by metal
    if metal is not None:
        df = df[df["element0"] == metal]

    # Sort dataframe based on sortBy parameter (q4-6 values)
    df = df.sort_values([sortBy], ascending = False).dropna(subset=[sortBy])
    if sortBy in ['q4','q5']:
        q = 'q' + str(int(sortBy[-1]) + 1)
        df = df[df[q] != np.nan]

    i_widget = IntSlider(
        min=0, max=df.shape[0] - 1, description='Structure', continuous_update=False)

    def get_neighbors_chain(i):
        return [df[f'chain{j}'].iloc[i] for j in range(1, 7) if df[f'element{j}'] is not None]

    def get_neighbors_group(i):
        return [df[f'groupNum{j}'].iloc[i] for j in range(1, 7) if df[f'element{j}'] is not None]

    def get_neighbors_elements(i):
        elements = [df[f'element{j}'].iloc[i]
                    for j in range(1, 7) if df[f'element{j}'] is not None]
        return [str(e).upper() for e in elements]

    def view3d(i=0):
        '''Simple structure viewer that uses py3Dmol to view PDB structure by
        indexing the list of PDBids
        Attributes
        ----------
        i (int): index of the protein if a list of PDBids
        '''

        structures = df['pdbId'].iloc
        groups = df['groupNum0'].iloc
        chains = df['chain0'].iloc
        elements = df['element0'].iloc
        ori = str(df[sortBy].iloc[i])[:5]

        print(f"PDBId: {structures[i]}    chain: {chains[i]}    element:    {elements[i]}")
        print(f"{sortBy}: {ori}")

        viewer = py3Dmol.view(query='pdb:' + structures[i], width=700, height=700)
        neighbors = {'resi': get_neighbors_group(i), 'chain': get_neighbors_chain(i)}
        metal = {'resi': groups[i], 'atom': str(elements[i]).upper(), 'chain': chains[i]}

        viewer.setStyle(neighbors, {'stick': {'colorscheme': 'orangeCarbon'}})
        viewer.setStyle(metal, {'sphere': {'radius': 0.5, 'color': 'gray'}})

        viewer.zoomTo(neighbors)
        return viewer.show()

    return interact(view3d, i=i_widget)


def metal_distance_widget(df_concat):
    '''Plot an violinplot of metal-element distances with ipywidgets

    Attributes
    ----------
        df_concat (Dataframe): dataframe of metal-elements distances
    '''
    metals = df_concat['Metal'].unique().tolist()
    m_widget = Dropdown(options = metals, description = "Metals")

    def metal_distance_violinplot(metal):
        df_metal = df_concat[df_concat["Metal"] == metal].copy()
        df_metal['Element'] = df_metal['Element'].apply(lambda x: metal+"-"+x)

        # Set fonts
        fig, ax = plt.subplots()
        fig.set_size_inches(15,6)
        subplot = sns.violinplot(x="Element", y="Distance", palette="muted", data=df_metal, ax=ax)
        subplot.set(xlabel="Metal Interactions", ylabel="Distance", title=f"{metal} to Elements Distances Violin Plot")

    return interact(metal_distance_violinplot, metal=m_widget);



#!/user/bin/env python
'''resolution.py:

This filter returns true if the resolution value for this
structure is within the sepcified range

References
----------
    Resolution:
        http://pdb101.rcsb.org/learn/guide-to-understanding-pdb-data/resolution

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''


class Resolution(object):
    '''This filter returns true if the resolution value of this structure is
    withing the specified range.

    Attributes
    ----------
        min_resolution (float): The lower bound resolution
        max_resolution (float): The upper bound resolution
    '''

    def __init__(self, minResolution, maxResolution):
        self.min_Resolution = minResolution
        self.max_Resolution = maxResolution

    def __call__(self, t):
        if t[1].resolution == None:
            return False

        return t[1].resolution >= self.min_Resolution and t[1].resolution <= self.max_Resolution



#!/user/bin/env python
'''structureToPolymerSequences.py:

This mapper maps a structure to it's polypeptides, polynucleotide chain sequences.
For a multi-model structure, only the first model is considered.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''


class StructureToPolymerSequences(object):
    '''This mapper maps a structure to it's polypeptides, polynucleotide chain sequences.
    For a multi-model structure, only the first model is considered.
    '''

    def __init__(self, useChainIdInsteadOfChainName=False, excludeDuplicates=False):
        '''Extracts all polymer chains from a structure. If the argument is set to true,
        the assigned key is: <PDB ID.Chain ID>, where Chain ID is the unique identifier
        assigned to each molecular entity in an mmCIF file. This Chain ID corresponds to
        <a href="http://mmcif.wwpdb.org/dictionaries/mmcif_mdb.dic/Items/_atom_site.label_asym_id.html">
        _atom_site.label_asym_id</a> field in an mmCIF file.

        Attributes
        ----------
            useChainIdInsteadOfChainName if true, use the Chain Id in the key assignments
            excludeDuplicates if true, return only one chain for each unique sequence= t[1]
        '''
        self.useChainIdInsteadOfChainName = useChainIdInsteadOfChainName
        self.excludeDuplicates = excludeDuplicates

    def __call__(self, t):
        structure = t[1]
        sequences = list()
        seqSet = set()

        chainToEntityIndex = self._get_chain_to_entity_index(structure)

        for i in range(structure.chains_per_model[0]):
            polymer = structure.entity_list[chainToEntityIndex[i]]['type'] == 'polymer'

            if polymer:
                key = t[0]

                if '.' in key:
                    key = key.split('.')[0]

                key += '.'
                if self.useChainIdInsteadOfChainName:
                    key += structure.chain_id_list[i]
                else:
                    key += structure.chain_name_list[i]

                if self.excludeDuplicates:
                    if chainToEntityIndex[i] in seqSet:
                        continue
                    seqSet.add(chainToEntityIndex[i])

                sequences.append(
                    (key, structure.entity_list[chainToEntityIndex[i]]['sequence']))

        return sequences

    def _get_chain_to_entity_index(self, structure):
        entityChainIndex = [0] * structure.num_chains

        for i in range(len(structure.entity_list)):

            for j in structure.entity_list[i]['chainIndexList']:
                entityChainIndex[j] = i

        return entityChainIndex



#!/user/bin/env python
'''chemicalStructureQuery.py

This filter returns entries that contain groups with specified chemical structures (SMILES string).
This chemical structure query supports for query: exact, similar, substructure, and superstructure.
For details see references.

References:
    Chemical Structure Search:
        http://www.rcsb.org/pdb/staticHelp.do?p=help/advancedsearch/chemSmiles.html

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from mmtfPyspark.webfilters import AdvancedQuery


class ChemicalStructureQuery(object):

    EXACT = "Exact"
    SIMILAR = "Similar"
    SUBSTRUCTURE = "Substructure"
    SUPERSTRUCTURE = "Superstructure"

    def __init__(self, smiles, queryType="Substructure", percentSimilarity=0):
        '''Constructor to setup filter that matches any entry with at least one
        chemical component that matches the specified SMILES string using the
        specified query type.

        For details see:
            <a href="http://www.rcsb.org/pdb/staticHelp.do?p=help/advancedsearch/chemSmiles.html">Chemical Structure Search</a>.

        Attribute
        ---------
            smiles (String): SMILES string representing chemical structure
            queryType: One of the 4 supported types
            percentSimilarity: percent similarity for similarity search. This parameter is
                               ignored for all other query types
        '''

        if not (queryType == self.EXACT
                or queryType == self.SIMILAR
                or queryType == self.SUBSTRUCTURE
                or queryType == self.SUPERSTRUCTURE):

            raise Exception("Invalid search type: %s" % queryType)

        query = "<orgPdbQuery>" + \
            "<queryType>org.pdb.query.simple.ChemSmilesQuery</queryType>" + \
            "<smiles>" + smiles + "</smiles>" + \
            "<searchType>" + queryType + "</searchType>" + \
            "<similarity>" + str(percentSimilarity) + "</similarity>" + \
            "<polymericType>Any</polymericType>" + \
            "</orgPdbQuery>"

        self.filter = AdvancedQuery(query)

    def __call__(self, t):
        return self.filter(t)



'''jpredDataset.py

This class downloads the dataset used to train the secondary structure predictor.
It can be used as a reference dataset for machine learning applications.

This dataset includes the ScopID, sequence, DSSP secondary structure assignment,
and a flag that indicates if data point was part of the training set.

Reference
---------
    JPred4:
        http://www.compbio.dundee.ac.uk/jpred/about_RETR_JNetv231_details.shtml

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

import urllib.request
import tarfile
from pyspark.sql import Row
from pyspark import SparkContext
from mmtfPyspark.ml import pythonRDDToDataset


def get_dataset():
    '''Gets JPred 4/JNet (v.2.3.1) secondary structure dataset.

    Return
    ------
        secondaryStructure dataset
    '''

    URL = "http://www.compbio.dundee.ac.uk/jpred/downloads/retr231.tar.gz"
    instream = urllib.request.urlopen(URL)
    secondaryStructures, sequences, trained = {}, {}, {}
    scopIds = set()
    res = []

    with tarfile.open(fileobj=instream, mode="r:gz") as tf:

        for entry in tf:
            if entry.isdir():
                continue
            br = tf.extractfile(entry)

            if ".dssp" in entry.name:
                scopID = str(br.readline())[3:-3]  # Remove newline and byte
                secondaryStructure = str(br.readline())[2:-3]  # Remove newline and byte
                secondaryStructure = secondaryStructure.replace('-', 'C')
                secondaryStructures[scopID] = secondaryStructure

            if ".fasta" in entry.name:
                scopID = str(br.readline())[3:-3]  # Remove newline and byte
                sequence = str(br.readline())[2:-3]  # Remove newline and byte
                scopIds.add(scopID)
                sequences[scopID] = sequence

                if "training/" in entry.name:
                    trained[scopID] = "true"
                elif "blind/" in entry.name:
                    trained[scopID] = "false"

    for scopId in scopIds:
        row = Row(scopId, sequences[scopId],
                  secondaryStructures[scopId], trained[scopId])
        res.append(row)

    sc = SparkContext.getOrCreate()
    data = sc.parallelize(res)
    colNames = ["scopID", "sequence", "secondaryStructure", "trained"]

    return pythonRDDToDataset.get_dataset(data, colNames)



'''proteinSequenceEncoder.py

This class encodes a protein sequence into a feature vector.
The protein sequence must be present in the input data set, the default column
name is "sequence". The default column name for the feature vector is "features".

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from pyspark.sql import SparkSession
from pyspark.ml.linalg import Vectors
from pyspark.ml.linalg import VectorUDT
from pyspark.ml.feature import Word2Vec, Word2VecModel
from mmtfPyspark.ml import sequenceNgrammer


class ProteinSequenceEncoder(object):
    '''
    This class encodes a protein sequence into a feature vector.
    The protein sequence must be present in the input data set,
    the default column name is "sequence". The default column name
    for the feature vector is "features".

    Attributes
    ----------
        data (DataFrame): input data to be encoded [None]
        inputCol (str): name of the input column [sequence]
        outputCol (str): name of the output column [features]
    '''

    model = None

    AMINO_ACIDS21 = ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', \
                     'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'X', 'Y']

    properties = {
        'A' : [1.28,0.05,1.00,0.31,6.11,0.42,0.23],
        'G' : [0.00,0.00,0.00,0.00,6.07,0.13,0.15],
        'V' : [3.67,0.14,3.00,1.22,6.02,0.27,0.49],
        'L' : [2.59,0.19,4.00,1.70,6.04,0.39,0.31],
        'I' : [4.19,0.19,4.00,1.80,6.04,0.30,0.45],
        'F' : [2.94,0.29,5.89,1.79,5.67,0.30,0.38],
        'Y' : [2.94,0.30,6.47,0.96,5.66,0.25,0.41],
        'W' : [3.21,0.41,8.08,2.25,5.94,0.32,0.42],
        'T' : [3.03,0.11,2.60,0.26,5.60,0.21,0.36],
        'S' : [1.31,0.06,1.60,-0.04,5.70,0.20,0.28],
        'R' : [2.34,0.29,6.13,-1.01,10.74,0.36,0.25],
        'K' : [1.89,0.22,4.77,-0.99,9.99,0.32,0.27],
        'H' : [2.99,0.23,4.66,0.13,7.69,0.27,0.30],
        'D' : [1.60,0.11,2.78,-0.77,2.95,0.25,0.20],
        'E' : [1.56,0.15,3.78,-0.64,3.09,0.42,0.21],
        'N' : [1.60,0.13,2.95,-0.60,6.52,0.21,0.22],
        'Q' : [1.56,0.18,3.95,-0.22,5.65,0.36,0.25],
        'M' : [2.35,0.22,4.43,1.23,5.71,0.38,0.32],
        'P' : [2.67,0.00,2.72,0.72,6.80,0.13,0.34],
        'C' : [1.77,0.13,2.43,1.54,6.35,0.17,0.41],
        'X' : [0.00,0.00,0.00,0.00,0.00,0.00,0.00],
    }

	# Source: https://ftp.ncbi.nih.gov/repository/blocks/unix/blosum/BLOSUM/blosum62.blast.new
    blosum62 = {
         #      A  R  N  D  C  Q  E  G  H  I  L  K  M  F  P  S  T  W  Y  V
        'A' : [ 4,-1,-2,-2, 0,-1,-1, 0,-2,-1,-1,-1,-1,-2,-1, 1, 0,-3,-2, 0],
        'R' : [-1, 5, 0,-2,-3, 1, 0,-2, 0,-3,-2, 2,-1,-3,-2,-1,-1,-3,-2,-3],
        'N' : [-2, 0, 6, 1,-3, 0, 0, 0, 1,-3,-3, 0,-2,-3,-2, 1, 0,-4,-2,-3],
        'D' : [-2,-2, 1, 6,-3, 0, 2,-1,-1,-3,-4,-1,-3,-3,-1, 0,-1,-4,-3,-3],
        'C' : [ 0,-3,-3,-3, 9,-3,-4,-3,-3,-1,-1,-3,-1,-2,-3,-1,-1,-2,-2,-1],
        'Q' : [-1, 1, 0, 0,-3, 5, 2,-2, 0,-3,-2, 1, 0,-3,-1, 0,-1,-2,-1,-2],
        'E' : [-1, 0, 0, 2,-4, 2, 5,-2, 0,-3,-3, 1,-2,-3,-1, 0,-1,-3,-2,-2],
        'G' : [ 0,-2, 0,-1,-3,-2,-2, 6,-2,-4,-4,-2,-3,-3,-2, 0,-2,-2,-3,-3],
        'H' : [-2, 0, 1,-1,-3, 0, 0,-2, 8,-3,-3,-1,-2,-1,-2,-1,-2,-2, 2,-3],
        'I' : [-1,-3,-3,-3,-1,-3,-3,-4,-3, 4, 2,-3, 1, 0,-3,-2,-1,-3,-1, 3],
        'L' : [-1,-2,-3,-4,-1,-2,-3,-4,-3, 2, 4,-2, 2, 0,-3,-2,-1,-2,-1, 1],
        'K' : [-1, 2, 0,-1,-3, 1, 1,-2,-1,-3,-2, 5,-1,-3,-1, 0,-1,-3,-2,-2],
        'M' : [-1,-1,-2,-3,-1, 0,-2,-3,-2, 1, 2,-1, 5, 0,-2,-1,-1,-1,-1, 1],
        'F' : [-2,-3,-3,-3,-2,-3,-3,-3,-1, 0, 0,-3, 0, 6,-4,-2,-2, 1, 3,-1],
        'P' : [-1,-2,-2,-1,-3,-1,-1,-2,-2,-3,-3,-1,-2,-4, 7,-1,-1,-4,-3,-2],
        'S' : [ 1,-1, 1, 0,-1, 0, 0, 0,-1,-2,-2, 0,-1,-2,-1, 4, 1,-3,-2,-2],
        'T' : [ 0,-1, 0,-1,-1,-1,-1,-2,-2,-1,-1,-1,-1,-2,-1, 1, 5,-2,-2, 0],
        'W' : [-3,-3,-4,-4,-2,-2,-3,-2,-2,-3,-2,-3,-1, 1,-4,-3,-2,11, 2,-3],
        'Y' : [-2,-2,-2,-3,-2,-1,-2,-3, 2,-1,-1,-2,-1, 3,-3,-2,-2, 2, 7,-1],
        'V' : [ 0,-3,-3,-3,-1,-2,-2,-3,-3, 3, 1,-2, 1,-1,-2,-2, 0,-3,-1, 4],
        'X' : [-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4],
    }


    def __init__(self, data = None, inputCol = "sequence", outputCol = "features"):

        self.data = data
        self.inputCol = inputCol
        self.outputCol = outputCol


    def one_hot_encode(self, data = None, inputCol = None, outputCol = None):
        '''
        One-hot encodes a protein sequence. The one-hot encoding
        encodes the 20 natural amino acids, plus X for any other
        residue for a total of 21 elements per residue.

        Attributes
        ----------
            data (DataFrame): input data to be encoded [None]
            inputCol (str): name of the input column [None]
            outputCol (str): name of the output column [None]
        '''

        # Setting class variables
        if data is not None:
            self.data = data

        if inputCol is not None:
            self.inputCol = inputCol

        if outputCol is not None:
            self.outputCol = outputCol

        if self.data is None:
            raise ValueError("Class variable data is not defined, please pass\
                             in a dataframe into the data parameter")

        session = SparkSession.builder.getOrCreate()
        AMINO_ACIDS21 = self.AMINO_ACIDS21

        # Encoder function to be passed as User Defined Function (UDF)
        def _encoder(s):

            values = [0] * len(AMINO_ACIDS21) * len(s)

            for i in range(len(s)):

                if s[i] in AMINO_ACIDS21:
                    index = AMINO_ACIDS21.index(s[i])

                else:
                    index = AMINO_ACIDS21.index('X')

                values[i*len(AMINO_ACIDS21) + index] = 1

            return Vectors.dense(values)

        session.udf.register("encoder", _encoder, VectorUDT())

        self.data.createOrReplaceTempView("table")
        sql = f"SELECT *, encoder({self.inputCol}) AS {self.outputCol} from table"

        data = session.sql(sql)

        return data


    def property_encode(self, data = None, inputCol = None, outputCol = None):
        '''Encodes a protein sequence by 7 physicochemical properties

        Reference
        ---------
            Meiler, J., Müller, M., Zeidler, A. et al. J Mol Model (2001)
    	    https://link.springer.com/article/10.1007/s008940100038

        Attributes
        ----------
            data (DataFrame): input data to be encoded [None]
            inputCol (str): name of the input column [None]
            outputCol (str): name of the output column [None]

        Returns
        -------
            dataset with feature vector appended
        '''

        # Setting class variables
        if data is not None:
            self.data = data

        if inputCol is not None:
            self.inputCol = inputCol

        if outputCol is not None:
            self.outputCol = outputCol

        if self.data is None:
            raise ValueError("Class variable data is not defined, please pass\
                             in a dataframe into the data parameter")

        session = SparkSession.builder.getOrCreate()
        properties = self.properties

        #Encoder function to be passed as User Defined Function (UDF)
        def _encoder(s):
            values = []

            for i in range(len(s)):

                if s[i] in properties:
                    values += properties[s[i]]

            return Vectors.dense(values)

        session.udf.register("encoder", _encoder, VectorUDT())

        self.data.createOrReplaceTempView("table")
        sql = f"SELECT *, encoder({self.inputCol}) AS {self.outputCol} from table"

        data = session.sql(sql)

        return data


    def blosum62_encode(self, data = None, inputCol = None, outputCol = None):
        '''Encodes a protein sequence by 7 Blosum62

        Reference
        ---------
            Blosum Matrix
            https://ftp.ncbi.nih.gov/repository/blocks/unix/blosum/BLOSUM/blosum62.blast.new

        Attributes
        ----------
            data (DataFrame): input data to be encoded [None]
            inputCol (str): name of the input column [None]
            outputCol (str): name of the output column [None]

        Returns
        -------
            dataset with feature vector appended
        '''

        if data is not None:
            self.data = data

        if inputCol is not None:
            self.inputCol = inputCol

        if outputCol is not None:
            self.outputCol = outputCol

        if self.data is None:
            raise ValueError("Class variable data is not defined, please pass\
                             in a dataframe into the data parameter")

        session = SparkSession.builder.getOrCreate()
        blosum62 = self.blosum62

        #Encoder function to be passed as User Defined Function (UDF)
        def _encoder(s):
            values = []

            for i in range(len(s)):

                if s[i] in blosum62:
                    values += blosum62[s[i]]

            return Vectors.dense(values)

        session.udf.register("encoder", _encoder, VectorUDT())

        self.data.createOrReplaceTempView("table")
        sql = f"SELECT *, encoder({self.inputCol}) AS {self.outputCol} from table"

        data = session.sql(sql)

        return data


    def overlapping_ngram_word2vec_encode(self, data = None, inputCol = None,
                                          outputCol = None, n = None,
                                          windowSize = None, vectorSize = None,
                                          fileName = None, sc = None):
        '''Encodes a protein sequence by converting it into n-grams and
        then transforming it into a Word2Vec feature vector.

        If given word2Vec file name, then this function encodes a protein
        sequence by converting it into n-grams and then transforming it using
        pre-trained word2Vec model read from that file

        Attribute
        ---------
            data (DataFrame): input data to be encoded [None]
            inputCol (str): name of the input column [None]
            outputCol (str): name of the output column [None]
            n (int): The number of words in an n-gram [None]
            windowSize (int): width of the window used to slide across the \
                              squence, context words from -window to window \
                              [None]
            vectorSize (int): dimension of the feature vector [None]
            fileName (str): filename of Word2Vec model [None]

        Returns
        -------
            dataset with features vector added to original dataset
        '''

        if data is not None:
            self.data = data

        if inputCol is not None:
            self.inputCol = inputCol

        if outputCol is not None:
            self.outputCol = outputCol

        if self.data is None:
            raise ValueError("Class variable data is not defined, please pass\
                             in a dataframe into the data parameter")

        # Create n-grams out of the sequence
        # E.g., 2-gram IDCGH, ... =>[ID, DC, CG, GH, ...]

        data = sequenceNgrammer.ngram(self.data, n, "ngram")

        if not (n == None and windowSize == None and vectorSize == None):
            # Convert n-grams to W2V freature vector
            # [ID, DC, CG, GH, ...] => [0.1234, 0.2394, ...]
            word2Vec = Word2Vec()
            word2Vec.setInputCol("ngram") \
                    .setOutputCol(self.outputCol) \
                    .setNumPartitions(8) \
                    .setWindowSize(windowSize) \
                    .setVectorSize(vectorSize) \

            self.model = word2Vec.fit(data)

        elif fileName != None and sc != None:
            reader = Word2VecModel()

            self.model = reader.load(sc, fileName)

            print(f"model file : {fileName} \n \
                    inputCol : {self.model.getInputCol()} \n \
                    windowSize : {self.model.getWindowSize()} \n \
                    vectorSize : {self.model.getVectorSize()}")

            self.model.setOutputCol(self.outputCol)

        else:
            raise Exception("Either provide word2Vec file (filename) + SparkContext (sc), \
                            or number of words(n) + window size(windowSize) \
                            + vector size (vetorSize), for function parameters")
            return

        return self.model.transform(data)


    def shifted_3gram_word2vec_encode(self, data = None, inputCol = None,
                                      outputCol = None, windowSize = None,
                                      vectorSize = None, fileName = None,
                                      sc = None):
        '''Encodes a protein sequence as three non-overlapping 3-grams,
        trains a Word2Vec model on the 3-grams, and then averages the
        three resulting freature vectors.


        Attribute
        ---------
            data (DataFrame): input data to be encoded [None]
            inputCol (str): name of the input column [None]
            outputCol (str): name of the output column [None]
            windowSize (int): width of the window used to slide across the sequence
                              context words from -window to window
            vectorSize (int): dimension of the feature vector [None]
            fileName (string): filename of Word2VecModel [None]
            sc (SparkContext): spark context [None]

        Returns
        -------
            dataset with features vector added to original dataset

        References
        ----------
           Asgari E, Mofrad MRK (2015) Continuous Distributed Representation
           of Biological Sequences for Deep Proteomics and Genomics.
           PLOS ONE 10(11): e0141287. doi:
           https://doi.org/10.1371/journal.pone.0141287
        '''

        if data is not None:
            self.data = data

        if inputCol is not None:
            self.inputCol = inputCol

        if outputCol is not None:
            self.outputCol = outputCol

        if self.data is None:
            raise ValueError("Class variable data is not defined, please pass\
                             in a dataframe into the data parameter")

        # Create n-grams out of the sequence
        # e.g., 2-gram [IDCGH, ...] => [ID. DC, CG, GH,...]

        data = sequenceNgrammer.shifted_ngram(self.data, 3, 0, "ngram0")
        data = sequenceNgrammer.shifted_ngram(data, 3, 1, "ngram1")
        data = sequenceNgrammer.shifted_ngram(data, 3, 2, "ngram2")
        if not (windowSize == None and vectorSize == None):

            ngram0 = data.select("ngram0").withColumnRenamed("ngram0","ngram")
            ngram1 = data.select("ngram1").withColumnRenamed("ngram1","ngram")
            ngram2 = data.select("ngram2").withColumnRenamed("ngram2","ngram")

            ngrams = ngram0.union(ngram1).union(ngram2)

            # Convert n-grams to W2V feature vector
            # [I D, D C, C G, G H, ... ] => [0.1234, 0.2394, .. ]
            word2Vec = Word2Vec()

            word2Vec.setInputCol("ngram") \
                    .setOutputCol("feature") \
                    .setMinCount(10) \
                    .setNumPartitions(8) \
                    .setWindowSize(windowSize) \
                    .setVectorSize(vectorSize)

            self.model = word2Vec.fit(ngrams)

        elif fileName != None and sc != None:
            reader = Word2VecModel()

            self.model = reader.load(sc, fileName)

            print(f"model file : {fileName} \n \
                    inputCol : {self.model.getInputCol()} \n \
                    windowSize : {self.model.getWindowSize()} \n \
                    vectorSize : {self.model.getVectorSize()}")

        else:
            raise Exception("Either provide word2Vec file (filename) + SparkContext (sc), \
                            or window size(windowSize) + vector size (vetorSize), \
                            for function parameters")
            return

        #data = data.withColumn("feature0",self.model.transform(data.select('ngram0').withColumnRenamed("ngram0","ngram")))
        for i in reversed(range(3)):
            feature = self.model.transform(data.select('ngram' + str(i)).withColumnRenamed("ngram" + str(i),"ngram"))
            data = data.join(feature.withColumnRenamed("ngram","ngram" + str(i)), "ngram" + str(i))
            data = data.withColumnRenamed("feature", "feature" + str(i))


        data = self._average_feature_vectors(data, self.outputCol)
        data.printSchema()

        cols = ['structureChainId','sequence','labelQ8','labelQ3','ngram0','ngram1',\
                'ngram2','feature0','feature1','feature2', 'features']

        data = data.select(cols)


        return data


    def get_word2vec_model(self):
        '''Returns a Word2VecModel created by overlapping_ngram_word2vec_encode()

        Returns
        -------
            overlapping Ngram Word2VecModel if available, otherwise None
        '''

        return self.model


    def _average_feature_vectors(self, data, outputCol):
        '''Average the feature vectors

        Attributes
        ----------
            data (DataFrame): input dataframe
            outputCol (str): name of the output column
        '''

        session = SparkSession.builder.getOrCreate()

        def _averager(v1, v2, v3):
            f1 = v1.toArray()
            f2 = v2.toArray()
            f3 = v3.toArray()

            length = min(len(f1), len(f2), len(f3))
            average = []

            for i in range(length):
                average.append((f1[i] + f2[i] + f3[i])/3.0)

            return Vectors.dense(average)

        session.udf.register("averager", _averager, VectorUDT())

        data.createOrReplaceTempView("table")

        sql = f"SELECT *, averager(feature0, feature1, feature2) AS {self.outputCol} from table"

        data = session.sql(sql)

        return data



#!/user/bin/env python
'''interactionCenter.py:

InteractionCenter stores information about an atom involved in a molecular
interaction.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''

from pyspark.sql.types import *
import numpy as np


class InteractionCenter(object):
    '''Class that stores information about an atom involved in a molecular
    interaction.

    Attributes
    ----------
        structure (mmtfStructure): structure to be used as interaction center [None]
        atomIndex (int): the index of the atom at center [None]
    '''

    def __init__(self, structure=None, atomIndex=None):

        if structure is not None and atomIndex is not None:

            self.set_atom_name(structure.get_atom_names()[atomIndex])
            self.set_element(structure.get_elements()[atomIndex])
            self.set_group_name(structure.get_group_names()[atomIndex])
            self.set_group_number(structure.get_group_numbers()[atomIndex])
            self.set_type(structure.get_entity_types()[atomIndex])
            self.set_chain_name(structure.get_chain_names()[atomIndex])
            self.set_sequence_position(
                structure.get_sequence_positions()[atomIndex])
            self.set_coordinates(np.array([structure.get_x_coords()[atomIndex],
                                           structure.get_y_coords()[atomIndex],
                                           structure.get_z_coords()[atomIndex]]))
            self.set_normalized_b_factors(
                structure.get_normalized_b_factors()[atomIndex])

        else:
            self.coordinates = None
            self.atomName = None
            self.element = None
            self.groupName = None
            self.groupNumber = None
            self.type = None
            self.chainName = None
            self.normalizedbFactor = None

    def get_length():
        '''Returns the number of data items in an interaction center.
        Note, not all data are currently included

        Returns
        -------
            the number of data items in an interaction center
        '''

        LENGTH = 7

        return LENGTH

    def get_atom_name(self):
        '''Gets the atom name

        Returns
        -------
            name of the atom
        '''

        return self.atomName

    def set_atom_name(self, atomName):
        '''Sets the atom name

        Attribute
        ---------
            atomName (str): name of the atom
        '''

        self.atomName = atomName

    def get_element(self):
        '''Gets the case-sensitive element symbol

        Returns
        -------
            element symbol
        '''

        return self.element

    def set_element(self, element):
        '''Sets the case-sensitive element symbol

        Attributes
        ----------
            element (str): element symbol
        '''

        self.element = element

    def get_group_name(self):
        '''Gets the names of the group. This name is the chemical component id
        of this group

        Returns
        -------
            name of group
        '''

        return self.groupName

    def set_group_name(self, groupName):
        '''Sets the names of the group. This name is the chemical component id
        of this group

        Attributes
        ----------
            groupName (str): name of group
        '''

        self.groupName = groupName

    def get_group_number(self):
        '''Gets the group number for the interaction center. A group number
        consists of the residue number (e.g. 101) plus an optional insertion code
        (e.g. A): 101A.

        Returns
        -------
            group number
        '''

        return self.groupNumber

    def set_group_number(self, groupNumber):
        '''Sets the group number for the interaction center. A group number
        consists of the residue number (e.g. 101) plus an optional insertion code
        (e.g. A): 101A.

        Attributes
        ----------
            groupNumber (str): group number
        '''

        self.groupNumber = groupNumber

    def get_type(self):
        '''Gets the type of the group.

        Returns
        -------
            type of group
        '''

        return self.type

    def set_type(self, gtype):
        '''Sets the type of the group.

        Attributes
        ----------
            gtype (str): type of group
        '''

        self.type = gtype

    def get_chain_name(self):
        '''Gets the chainName. This corresponds to the "chain Id" in PDB files.

        Returns
        -------
            the name of the chain
        '''

        return self.chainName

    def set_chain_name(self, chainName):
        '''Sets the chainName. This corresponds to the "chain Id" in PDB files.

        Attributes
        ----------
            chainName (str): the name of the chian
        '''

        self.chainName = chainName

    def get_sequence_position(self):
        '''Gets an index into the one-letter polymer sequence. this index is
        zero-based. If the interaction center is not a polymer atom, this index
        is -1

        Returns
        -------
            index into polymer sequence
        '''

        return self.sequencePosition

    def set_sequence_position(self, sequencePosition):
        '''Sets an index into the one-letter polymer sequence. this index is
        zero-based. If the interaction center is not a polymer atom, this index
        is -1

        Attributes
        ----------
            sequencePosition (int): index into polymer sequence
        '''

        self.sequencePosition = sequencePosition

    def get_coordinates(self):
        '''Gets the position of the interaction center.

        Returns
        -------
            the position of the interaction center
        '''

        return self.coordinates

    def set_coordinates(self, coordinates):
        '''Sets the position of the interaction center.

        Attributes
        ----------
            coordinates (list): the position of the interaction center
        '''

        self.coordinates = coordinates

    def get_normalized_b_factors(self):
        '''Gets the normalized b-factor

        Returns
        -------
            the normalized b-factor
        '''

        return self.normalizedbFactor

    def set_normalized_b_factors(self, normalizedbFactor):
        '''Sets the normalized b-factor

        Attributes
        ----------
            normalizedbFactor (float) the normalized b-factor
        '''

        self.normalizedbFactor = normalizedbFactor

    def get_as_object(self):
        '''Returns a list of objects representing this interaction center. Note,
        not all data are currently included. This method is ucsed to create
        Spark datasets.

        Returns
        -------
            list of objects representing this interaction center
        '''

        return [self.atomName, self.element, self.groupName, self.groupNumber,
                self.type, self.chainName, self.normalizedbFactor]

    def get_struct_fields(index):
        '''Returns a schema to create Spark Datasets. This schema must match the
        order in which the data are return by the getAsObject() method.

        Attributes
        ----------
            index (int): an integer to label an interaction center

        Returns
        -------
            schema to represent an interaction center in spark dataset.
        '''

        sf = []
        nullable = True
        index = str(index)

        sf.append(StructField("atom" + index, StringType(), nullable))
        sf.append(StructField("element" + index, StringType(), nullable))
        sf.append(StructField("group" + index, StringType(), nullable))
        sf.append(StructField("groupNum" + index, StringType(), nullable))
        sf.append(StructField("type" + index, StringType(), nullable))
        sf.append(StructField("chain" + index, StringType(), nullable))
        sf.append(StructField("nbFactor" + index, FloatType(), nullable))

        return sf



from setuptools import setup

mmtfPyspark_packages = ['mmtfPyspark',
                        'mmtfPyspark.datasets',
                        'mmtfPyspark.filters',
                        'mmtfPyspark.io',
                        'mmtfPyspark.mappers',
                        'mmtfPyspark.ml',
                        'mmtfPyspark.utils',
                        'mmtfPyspark.webfilters',
                        'mmtfPyspark.webservices',
                        'mmtfPyspark.interactions'
                        ]

mmtfPyspark_dependencies = ['pyspark>=2.3.0',
                            'Biopython>=1.7.1',
                            'msgpack-python>=0.5.6',
                            'numpy>=1.14.2',
                            'ipywidgets>=7.2.1',
                            'mmtf-python==1.1.1',
                            'sympy==1.1.1',
                            'requests>=0.8.0',
                            'matplotlib>=2.2.0',
                            'seaborn>=0.7.1',
                            'py3Dmol>=0.6.3',
                            'scipy>=1.0.0',
                            'scikit-learn>=0.13.0',
                            'py4j>=0.10.6'
                            ]

LONG_DESCRIPTION = """
**mmtfPyspark** is a python package that provides APIs and sample applications for distributed analysis and scalable mining of 3D biomacromolecular structures, such as the Protein Data Bank (PDB) archive. mmtfPyspark uses Big Data technologies to enable high-performance parallel processing of macromolecular structures.
"""

setup(name='mmtfPyspark',
      version='0.2.4',
      description='Methods for parallel and distributed analysis and mining of the Protein Data Bank using MMTF and Apache Spark',
      long_description=LONG_DESCRIPTION,
      url='https://github.com/sbl-sdsc/mmtf-pyspark',
      author='Mars Huang (Shih-Cheng)',
      author_email='marshuang80@gmail.com',
      license='Apache License 2.0',
      keywords='mmtf spark pyspark protein PDB',
      packages=mmtfPyspark_packages,
      install_requires=mmtfPyspark_dependencies,
      python_requires='>=3.6',
      include_package_data=True,
      test_suite='nose.collector',
      test_require=['nose'],
      classifiers=['Development Status :: 3 - Alpha',
                 'Intended Audience :: Science/Research',
                 'Topic :: Scientific/Engineering :: Bio-Informatics',
                 'License :: OSI Approved :: Apache Software License',
                 'Programming Language :: Python :: 3.6'],
      zip_safe=False)



#!/user/bin/env python
'''wildType.py

This filter returns entries that contain wild type protein chains.
polymer chain(s) made of L-amino acids. If the "exclusive" flag is set to true
in the constructor, all polymer chains must be L-proteins. For a multi-model structure,
this filter only checks the first model.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from mmtfPyspark.webfilters import AdvancedQuery


class WildTypeQuery(object):

    SEQUENCE_COVERAGE_100 = 100
    SEQUENCE_COVERAGE_95 = 95
    SEQUENCE_COVERAGE_90 = 90
    SEQUENCE_COVERAGE_85 = 85
    SEQUENCE_COVERAGE_80 = 80
    SEQUENCE_COVERAGE_75 = 75
    SEQUENCE_COVERAGE_70 = 70
    SEQUENCE_COVERAGE_65 = 65
    SEQUENCE_COVERAGE_60 = 60

    def __init__(self, includeExpressionTags, percentSequenceCoverage=None):
        '''Default constructor maches an entry that contains at least one L-protein chain.
        As an example, an L-protein/DNA complex passes this filter

        Attributes
        ----------
            includeExpressionTags (bool): flag to include expression tags
            percentSequenceCoverage (int): percentage of sequence converage [NONE]
        '''

        query = "<orgPdbQuery><queryType>org.pdb.query.simple.WildTypeProteinQuery</queryType>"

        if includeExpressionTags:
            query += "<includeExprTag>Y</includeExprTag>"
        else:
            query += "<includeExprTag>N</includeExprTag>"

        if not percentSequenceCoverage == None:
            query += "<percentSeqAlignment>%i</percentSeqAlignment>"%percentSequenceCoverage

        query += "</orgPdbQuery>"

        self.filter = AdvancedQuery(query)

    def __call__(self, t):
        return self.filter(t)



#!/user/bin/env python
'''advancedQueryService.py

Post an XML query (PDB XML query format) to the RESTful RCSB web service

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

import urllib

SERVICELOCATION = "http://www.rcsb.org/pdb/rest/search"


def post_query(xml):
    '''Post an XML query (PDB XML query format) to the RESTful
    RCSB web service

    Attribute
    ---------
        xml (str): a string of xml query
    '''

    encodedXML = urllib.parse.quote(xml).encode('utf-8')

    url = urllib.request.Request(SERVICELOCATION)

    with urllib.request.urlopen(url, data=encodedXML) as f:

        pdbIds = [str(l)[2:-3] for l in f.readlines()]

    return pdbIds



from .pisces import Pisces
from .advancedQuery import AdvancedQuery
from .chemicalStructureQuery import ChemicalStructureQuery
from .sequenceSimilarity import SequenceSimilarity
from .wildTypeQuery import WildTypeQuery
from .customReportQuery import CustomReportQuery
from .blastCluster import BlastCluster
from .pdbjMineSearch import PdbjMineSearch



#!/user/bin/env python
'''columnarStructure.py

Provides efficient access to structure information in the form of atom-based arrays.
Data are lazily initialized as needs.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

import numpy as np


class ColumnarStructure(object):
    '''Column based data structure to efficiently access structure information

    Attributes
    ----------
        structure (mmtfStructure): mmtf structure
        firstModelOnly (bool): use only the first model in a structure if True
    '''

    def __init__(self, structure, firstModelOnly=True):

        # Set class variables
        self.numAtoms = 0
        self.numChains = 0
        self.numGroups = 0
        self.numModels = 0

        self.atomNames = None
        self.atomToChainIndices = None
        self.atomToGroupIndices = None
        self.chainIds = None
        self.chainNames = None
        self.chemCompType = None
        self.elements = None
        self.entityTypes = None
        self.entityChainIndex = None
        self.entityIndices = None
        self.groupNames = None
        self.groupNumbers = None
        self.groupToAtomIndices = None
        self.polymer = None
        self.sequencePositions = None
        self.structure = structure
        self.groupToAtomIndices = None
        self.chainToAtomIndices = None
        self.chainToGroupIndices = None


        if firstModelOnly:
            self.numModels = 1
        else:
            self.numModels = structure.num_models

    def get_group_to_atom_indices(self):
        if self.groupToAtomIndices is None:
            self.get_indices()
        return self.groupToAtomIndices

    def get_chain_to_atom_indices(self):
        if self.chainToAtomIndices is None:
            self.get_indices()
        return self.chainToAtomIndices

    def get_chain_to_group_indices(self):
        if self.chainToGroupIndices is None:
            self.get_indices
        return self.chainToGroupIndices

    def get_num_atoms(self):
        self.get_indices()
        return self.numAtoms

    def get_num_groups(self):
        self.get_indices()
        return self.numGroups

    def get_num_chains(self):
        self.get_indices()
        return self.numChains

    def get_num_models(self):
        self.get_indices()
        return self.numModels

    def get_x_coords(self):
        return self.structure.x_coord_list

    def get_y_coords(self):
        return self.structure.y_coord_list

    def get_z_coords(self):
        return self.structure.z_coord_list

    def get_occupancies(self):
        return self.structure.occupancy_list

    def get_b_factors(self):
        return self.structure.b_factor_list

    def get_alt_loc_list(self):
        if not self.structure.alt_loc_set:
            self.structure = self.structure.set_alt_loc_list()
        return self.structure.alt_loc_list

    def get_group_types(self):
        return self.structure.group_type_list

    def get_atom_to_group_indices(self):

        if self.atomToGroupIndices is None:
            self.atomToGroupIndices = np.empty(
                self.get_num_atoms(), dtype='>i4')

            for i in range(self.get_num_groups()):
                start = self.groupToAtomIndices[i]
                end = self.groupToAtomIndices[i + 1]
                self.atomToGroupIndices[start:end] = i

        return self.atomToGroupIndices

    def get_chem_comp_types(self):

        if self.chemCompType is None:
            # Max chemCompType has 17 characters
            self.chemCompType = np.empty(
                self.get_num_atoms(), dtype=np.object_)
            self.groupTypeIndices = self.get_group_types()

            for i in range(self.get_num_groups()):
                index = self.groupTypeIndices[i]
                value = self.structure.group_list[index]['chemCompType']
                start = self.groupToAtomIndices[i]
                end = self.groupToAtomIndices[i + 1]
                self.chemCompType[start:end] = value

        return self.chemCompType

    def get_elements(self):

        if self.elements is None:
            self.elements = np.empty(self.get_num_atoms(), dtype=np.object_)
            self.groupTypeIndices = self.get_group_types()

            for i in range(self.get_num_groups()):
                index = self.groupTypeIndices[i]
                elementNames = self.structure.group_list[index]['elementList']
                start = self.groupToAtomIndices[i]
                end = self.groupToAtomIndices[i + 1]
                self.elements[start:end] = elementNames

        return self.elements

    def get_atom_names(self):

        if self.atomNames is None:
            # in case of 2 cjacter group + 3 digits
            self.atomNames = np.empty(self.get_num_atoms(), dtype=np.object_)
            self.groupTypeIndices = self.get_group_types()

            for i in range(self.get_num_groups()):
                index = self.groupTypeIndices[i]
                atomNames = self.structure.group_list[index]['atomNameList']
                start = self.groupToAtomIndices[i]
                end = self.groupToAtomIndices[i + 1]
                self.atomNames[start:end] = atomNames

        return self.atomNames

    def get_entity_types(self):

        if self.entityTypes is None:
            self.entityTypes = np.empty(self.get_num_atoms(), dtype=np.object_)

            # Instantiate required data
            self.is_polymer()
            self.get_chem_comp_types()
            self.get_elements()
            self.get_group_names()

            for i in range(self.get_num_groups()):
                start = self.groupToAtomIndices[i]
                end = self.groupToAtomIndices[i + 1]
                poly = self.polymer[start]
                ccType = self.chemCompType[start]

                if poly:
                    if "PEPTIDE" in ccType:
                        eType = "PRO"
                    elif "DNA" in ccType:
                        eType = "DNA"
                    elif "RNA" in ccType:
                        eType = "RNA"
                    elif "SACCHARIDE" in ccType:
                        eType = "PSR"
                    else:
                        eType = "UNK"
                elif (self.groupNames[start] == "HOH" or self.groupNames[start] == "DOD"):
                    eType = "WAT"
                elif "SACCHARIDE" in ccType:
                    eType = "SAC"
                else:
                    # If group contains at least one carbon atom, it is
                    # considered organic, but see expection below
                    organic = False
                    for j in range(start, end):
                        if self.elements[j] == "C":
                            organic = True
                            break

                    # Handle exceptions: carbon dioxide, carbon monoxide,
                    # cyanide ion are considered inorganic
                    if self.groupNames[start] == "CO2" or \
                       self.groupNames[start] == "CMO" or \
                       self.groupNames[start] == "CYN":
                        organic = False

                    if organic:
                        eType = "LGO"
                    else:
                        eType = "LGI"

                self.entityTypes[start:end] = eType

        return self.entityTypes

    def get_group_names(self):

        if self.groupNames is None:
            self.groupNames = np.empty(self.get_num_atoms(), dtype=np.object_)
            self.groupTypeIndices = self.structure.group_type_list

            for i in range(self.get_num_groups()):
                index = self.groupTypeIndices[i]
                value = self.structure.group_list[index]['groupName']
                start = self.groupToAtomIndices[i]
                end = self.groupToAtomIndices[i + 1]
                self.groupNames[start:end] = value

        return self.groupNames

    def get_group_numbers(self):

        if self.groupNumbers is None:
            # In case of 4 digit group id + 2 ins code
            self.groupNumbers = np.empty(
                self.get_num_atoms(), dtype=np.object_)

            for i in range(self.get_num_groups()):
                value = str(self.structure.group_id_list[i])
                insCode = self.structure.ins_code_list[i]

                if insCode != "\x00":
                    value += insCode

                start = self.groupToAtomIndices[i]
                end = self.groupToAtomIndices[i + 1]
                self.groupNumbers[start:end] = value

        return self.groupNumbers

    def get_chain_ids(self):

        if self.chainIds is None:
            self.chainIds = np.empty(self.get_num_atoms(), dtype=np.object_)

            for i in range(self.get_num_chains()):
                start = self.chainToAtomIndices[i]
                end = self.chainToAtomIndices[i + 1]
                self.chainIds[start:end] = self.structure.chain_id_list[i]

        return self.chainIds

    def get_chain_names(self):

        if self.chainNames is None:
            self.chainNames = np.empty(self.get_num_atoms(), dtype=np.object_)

            for i in range(self.get_num_chains()):
                start = self.chainToAtomIndices[i]
                end = self.chainToAtomIndices[i + 1]
                self.chainNames[start:end] = self.structure.chain_name_list[i]

        return self.chainNames

    def is_polymer(self):

        if self.polymer is None:
            # TODO: Numpy bool type is different than python bools, might need to fix
            self.polymer = np.empty(self.get_num_atoms(), dtype=bool)
            self.get_chain_to_entity_index()

            for i in range(self.get_num_chains()):
                index = self.entityChainIndex[i]
                start = self.chainToAtomIndices[i]
                end = self.chainToAtomIndices[i + 1]
                poly = self.structure.entity_list[index]['type'] == 'polymer'

                self.polymer[start:end] = poly

        return self.polymer

    def get_entity_indices(self):

        if self.entityIndices is None:
            self.entityIndices = np.empty(
                self.get_num_atoms(), dtype=np.object_)
            self.get_chain_to_entity_index()

            for i in range(self.get_num_chains()):
                start = self.chainToAtomIndices[i]
                end = self.chainToAtomIndices[i + 1]

                self.entityIndices[start:end] = self.entityChainIndex[i]

        return self.entityIndices

    def get_atom_to_chain_indices(self):

        if self.atomToChainIndices is None:
            self.atomToChainIndices = np.empty(
                self.get_num_atoms(), dtype='>i4')

            for i in range(self.get_num_chains()):
                start = self.chainToAtomIndices[i]
                end = self.chainToAtomIndices[i + 1]

                self.atomToChainIndices[start:end] = i

        return self.atomToChainIndices

    def get_sequence_positions(self):

        if self.sequencePositions is None:

            self.sequencePositions = np.empty(
                self.get_num_atoms(), dtype='>i4')
            groupSequenceIndices = self.structure.sequence_index_list

            for i in range(self.get_num_groups()):
                value = groupSequenceIndices[i]
                start = self.groupToAtomIndices[i]
                end = self.groupToAtomIndices[i + 1]

                self.sequencePositions[start:end] = value

        return self.sequencePositions

    def get_chain_to_entity_index(self):
        '''Returns an array that maps a chain index to an entity index

        Returns
        -------
            entityChainIndex [np.array] : index that maps chain index to an entity index
        '''

        if self.entityChainIndex is None:

            self.entityChainIndex = np.empty(
                self.structure.num_chains, dtype='>i4')

            for i, entity in enumerate(self.structure.entity_list):

                chainIndexList = entity['chainIndexList']
                self.entityChainIndex[chainIndexList] = i

        return self.entityChainIndex

    def get_indices(self):

        if self.groupToAtomIndices is None:

            self.groupToAtomIndices = np.empty(
                self.structure.num_groups + 1, dtype='>i4')
            self.chainToAtomIndices = np.empty(
                self.structure.num_chains + 1, dtype='>i4')
            self.chainToGroupIndices = np.empty(
                self.structure.num_chains + 1, dtype='>i4')

            chainCount, groupCount, atomCount = 0, 0, 0

            # Loop over all models
            for m in range(self.numModels):

                # Loop over all chains
                for i in range(self.structure.chains_per_model[m]):

                    self.chainToAtomIndices[chainCount] = atomCount
                    self.chainToGroupIndices[chainCount] = groupCount

                    # Loop over all groups in chain
                    for j in range(self.structure.groups_per_chain[i]):

                        groupType = self.structure.group_type_list[groupCount]
                        self.groupToAtomIndices[groupCount] = atomCount
                        atomsInGroup = len(
                            self.structure.group_list[groupType]['formalChargeList'])

                        groupCount += 1
                        atomCount += atomsInGroup

                    chainCount += 1

            self.groupToAtomIndices[groupCount] = atomCount
            self.chainToAtomIndices[chainCount] = atomCount
            self.chainToGroupIndices[chainCount] = groupCount

            self.numAtoms = atomCount
            self.numGroups = groupCount
            self.numChains = chainCount

            if self.numAtoms < self.structure.num_atoms:
                self.groupToAtomIndices = self.groupToAtomIndices[:groupCount + 1]
                self.chainToAtomIndices = self.chainToAtomIndices[:chainCount + 1]
                self.chainToGroupIndices = self.chainToGroupIndices[:chainCount + 1]



#!/user/bin/env python
'''structureToBiopython.py:

Maps a structure to BioPython Structure

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Error"
'''
from Bio.PDB.mmtf import DefaultParser
from mmtf.utils import decoder_utils

# TODO MMTFEncoder error


class StructureToBiopython(object):

    def __call__(self, t):

        parser = DefaultParser.StructureDecoder()
        t.pass_data_on(parser)

        return parser.structure_bulder.get_structure()



#!/user/bin/env python
'''drugBankDatset.py

This module provides access to DrugBank containing drug structure and drug target
imformation. These datasets contain identifiers and names for integration with
other data resources.

References
----------
    Drug Bank:
        https://www.drugbank.ca
    Wishart DS, et al., DrugBank 5.0: a major update to the DrugBank database
    for 2018. Nucleic Acids Res. 2017 Nov 8 :
        https://dx.doi.org/10.1093/nar/gkx1037

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

import requests
import tempfile
from zipfile import ZipFile
from io import BytesIO
from pyspark.sql import SparkSession

DRUG_GROUP = ['ALL', 'APPROVED', 'EXPERIMENTAL', 'NUTRACEUTICAL', 'ILLICIT'\
              'WITHDRAWN', 'INVESTIGATIONAL']
DRUG_TYPE = ['SMALL_MOLECULE', 'BIOTECH']
BASE_URL = "https://www.drugbank.ca/releases/latest/downloads/"
BUFFER = 2048


def get_open_drug_links():
    '''Downloads the DrugBank Open Data dataset with drug structure external
    links and identifiers. See DrugBank.

    This dataset contains drug common names, synonyms, CAS numbers, and
    Standard InChIKeys.

    The DrugBank Open Data dataset is a public domain dataset that can be
    used freely in your application or project (including commercial use).
    It is released under a Creative Common’s CC0 International License.

    References
    ----------
        Open Data dataset:
            https://www.drugbank.ca/releases/latest#open-data

    Example
    -------
        Get DrugBank open dataset:
            openDrugLinks = DrugBankDataset.get_open_drug_links()
            openDrugLinks.show()
        +----------+--------------------+-----------+--------------------+
        |DrugBankID|          Commonname|        CAS|    StandardInChIKey|
        +----------+--------------------+-----------+--------------------+
        |   DB00006|         Bivalirudin|128270-60-0|OIRCOABEOLEUMC-GE...|
        |   DB00014|           Goserelin| 65807-02-5|BLCLNMBMMGCOAS-UR...|

    Retruns
    -------
        DrugBank Dataset
    '''

    url = BASE_URL + "all-drugbank-vocabulary"
    return get_dataset(url)


def get_drug_links(drugGroup, username, password):
    '''Downloads drug structure external links and identifiers from DrugBank.
    Either all or subsets of data can be downloaded by specifying the
    DrugGroup:
        ALL, APPROVED, EXPERIMENTAL, NUTRACEUTICAL, ILLLICT, WITHDRAWN,
        INVESTIGATIONAL.

    The structure external links datasets include drug structure information
    in the form of InChI/InChI Key/SMILES as well as identifiers for other
    drug-structure resources (such as ChEBI, ChEMBL,ChemSpider, BindingDB,
    etc.). Included in each dataset is also the PubChem Compound ID (CID) and
    the particular PubChem Substance ID (SID) for the given DrugBank record.

    These DrugBank datasets are released under a Creative Common’s
    Attribution-NonCommercial 4.0 International License. They can be used
    freely in your non-commercial application or project. A DrugBank user
    account and authentication is required to download these datasets.

    References
    ----------
        External Drug Links:
            https://www.drugbank.ca/releases/latest#external-links

    Example
    -------
        Get dataset of external links and identifiers of approved drugs:
            username = "<your DrugBank username>"
            String password = "<your DrugBank password>"
            drugLinks = get_drug_links("APPROVED", username, password)
            drugLinks.show()

    Attributes
    ----------
        durgGroup (str): specific dataset to be downloaded, has to be in the
                         pre-defined DURG_GROUP list
        usesrname (str): DurgBank username
        password (str): DurgBank password

    Retruns
    -------
        DrugBank Dataset
    '''

    if drugGroup.upper() not in DRUG_GROUP:
        raise ValueError("drugGroup not in pre-defined durgGroups")

    url = BASE_URL + drugGroup + "-structure-links"
    return get_dataset(url, username, password)


def get_drug_target_links(drug, username, password):
    '''Downloads drug target external links and identifiers from DrugBank.
    Either all or subsets of data can be downloaded by specifying the
    DrugGroup:
        ALL, APPROVED, EXPERIMENTAL, NUTRACEUTICAL, ILLLICT, WITHDRAWN,
        INVESTIGATIONAL.
    OR DrugType:
        SMALL_MOLECULE, BIOTECH.

    The drug target external links datasets include drug name, drug type
    (small molecule, biotech), UniProtID and UniProtName.

    These DrugBank datasets are released under the Creative Common’s
    Attribution-NonCommercial 4.0 International License. They can be used
    freely in your non-commercial application or project. A DrugBank user
    account and authentication is required to download these datasets.

    References
    ----------
        Target Drug-UniProt:
            https://www.drugbank.ca/releases/latest#external-links

    Example
    -------
        Get dataset of drug target external links and identifiers of all drugs
        in DrugBank:
            username = "<your DrugBank username>"
            password = "<your DrugBank password>"
            drugTargetLinks = get_drug_target_links("ALL",
                                                    username,
                                                    password)
            drugTargetLinks.show()

    Attributes
    ----------
        durg (str): specific dataset to be downloaded, has to be either in
                    the DrugGroup list OR DrugType list.

        usesrname (str): DurgBank username
        password (str): DurgBank password

    Retruns
    -------
        DrugBank Dataset
    '''

    if drug.upper() in DRUG_GROUP or drug.upper() in DRUG_TYPE:
        url = BASE_URL + "target-" + drug + "-uniprot-links"
    else:
        raise ValueError("drug not in pre-defined durgGroups or drugTypes")

    return get_dataset(url, username, password)


def get_dataset(url, username=None, password=None):
    '''Downloads a DrugBank dataset

    Attributes
    ----------
        url (str): DrugBank dataset download links
        username (str): DrugBank username <None>
        password (str): DrugBank password <None>

    Returns
    -------
        DrugBank dataset
    '''
    if username is None and password is None:
        # get input stream to first zip entry
        req = requests.get(url)
    else:
        # TODO dataset that requires authentication
        req = requests.get(url, auth=(username, password))
        if req.text ==  'Invalid Email or password.':
                raise ValueError('Invalid Email or password.')

    # Decode and unzip file
    unzipped = _decode_as_zip_input_stream(req.content)

    # save data to a temporary file (Dataset csv reader requires a input
    # file!)
    tempFileName = _save_temp_file(unzipped)

    #load temporary CSV file to Spark dataste
    dataset = _read_csv(tempFileName)
    dataset = _remove_spaces_from_column_names(dataset)

    return dataset


def _decode_as_zip_input_stream(content):
    '''Returns an input stream to the first zip file entry

    Attributes
    ----------
        content (inputStream): inputStream content from request

    Returns
    -------
        unzipped InputStream
    '''

    zipfile = ZipFile(BytesIO(content))
    return [line.decode() for line \
            in zipfile.open(zipfile.namelist()[0]).readlines()]


def _save_temp_file(unzipped):
    '''Saves tabular report as a temporary CSV file

    Attributes
    ----------
        unzipped (list): list of unzipped content

    Returns
    -------
        path to the tempfile
    '''
    tempFile = tempfile.NamedTemporaryFile(delete=False)
    with open (tempFile.name, "w") as t:
        t.writelines(unzipped)
    return tempFile.name


def _read_csv(inputFileName):
    '''Reads CSV file into Spark dataset

    Attributes
    ----------
        fileName(str): name of the input csv fileName

    Returns
    -------
        a spark dataset
    '''
    spark = SparkSession.builder.getOrCreate()
    dataset = spark.read.format("csv") \
                        .option("header", "true") \
                        .option("inferSchema", "true") \
                        .load(inputFileName)
    return dataset


def _remove_spaces_from_column_names(original):
    '''Remove spaces from column names to ensure compatibility with parquet

    Attributes
    ----------
        original (Dataset): the original dataset

    Returns
    -------
        dataset with columns renamed
    '''

    for existingName in original.columns:
        newName = existingName.replace(' ','')
        # TODO: double check dataset "withColumnRenamed" funciton
        original = original.withColumnRenamed(existingName,newName)

    return original



#!/user/bin/env python
'''pisces.py

 This filter passes through representative structures or protein chains
 from the PISCES CulledPDB sets. A CulledPDB set is selected by specifying
 sequenceIdentity and resolution cutoff values from the following
 list:
    sequenceIdentity = [20, 25, 30, 40, 50, 60, 70, 80, 90]
    resolution = [1.6, 1.8, 2.0, 2.2, 2.5, 3.0]

References
----------
    PISCES:
        http://dunbrack.fccc.edu/PISCES.php
        Please cite the following in any work that uses lists provided by PISCES
        G. Wang and R. L. Dunbrack, Jr. PISCES: a protein sequence culling server.
        Bioinformatics, 19:1589-1591, 2003.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''
from mmtfPyspark.webservices import PiscesDownloader


class Pisces(object):
    '''Filters representative PDB structures and polymer chains based
     on the specified criteria using PISCES CulledPDB sets.
     <p> sequenceIdentity = 20, 25, 30, 40, 50, 60, 70, 80, 90
     <p> resolution = 1.6, 1.8, 2.0, 2.2, 2.5, 3.0

    Attributes
    ----------
        sequenceIdentity(int): sequence identity cutoff values
        resolution (float): resolution cutoff value
    '''

    def __init__(self, sequenceIdentity, resolution):
        self.pdbIds = set()

        pD = PiscesDownloader(sequenceIdentity, resolution)

        for pdbId in pD.get_structure_chain_ids():
            self.pdbIds.add(pdbId)
            self.pdbIds.add(pdbId[:4])

    def __call__(self, t):
        return t[0] in self.pdbIds



#!/user/bin/env python
'''secondaryStructureSegmentExtractor.py:

This class creates a dataset of sequence segments of specified length
and associate secondary structure information. Sequence and secondary
structure strings are split into segments using a sliding window of the specified
segment length. The dataset contains the sequence segment and the DSP Q8 and
DSSP Q3 secondary structure annotation of the cneter residue. Therefore, the segment
length must be an odd number

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from mmtfPyspark.datasets import secondaryStructureExtractor
from mmtfPyspark.mappers import StructureToSecondaryStructureSegments
from mmtfPyspark.ml import pythonRDDToDataset

def get_dataset(structureRDD, length):
    '''Returns a dataset of sequence segments of the specified length and
    the DSSP Q8 and Q3 code of the center residue in a segment.

    Attributes
    ----------
        structureRDD: structure
        length (int): segment length, must be an odd number

    Returns
    -------
        dataset of segments
    '''

    if length % 2 == 0:
        raise Exception("Segment length must be an odd number %i" % length)

    rows = secondaryStructureExtractor.get_python_rdd(structureRDD) \
            .flatMap(StructureToSecondaryStructureSegments(length))

    colNames = ["structureChainId", "sequence", "labelQ8", "labelQ3"]
    return pythonRDDToDataset.get_dataset(rows, colNames)
